'''
   Essential BC
'''
from petram.phys.numba_coefficient import func_to_numba_coeff_scalar
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.coefficient import VCoeff
from petram.mfem_config import use_parallel
import numpy as np

from petram.model import Bdry
from petram.phys.phys_model import Phys, VectorPhysCoefficient, PhysCoefficient
from petram.phys.em2da.em2da_base import EM2Da_Bdry, EM2Da_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2Da_E')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


data = (('E', VtableElement('E', type='complex',
                            guilabel='Electric field',
                            suffix=('r', 'phi', 'z'),
                            default=np.array([0, 0, 0]),
                            tip="essential BC")),)


class Ep(VectorPhysCoefficient):
    def EvalValue(self, x):
        v = super(Ep, self).EvalValue(x)
        v = np.array((v[0], v[2]))
        if self.real:
            return v.real
        else:
            return v.imag


class rEt(PhysCoefficient):
    def __init__(self, *args, **kwargs):
        #kwargs['isArray'] = True
        PhysCoefficient.__init__(self, *args, **kwargs)

    def EvalValue(self, x):
        v = super(rEt, self).EvalValue(x)
        v = v[1]*x[0]
        if self.real:
            return v.real
        else:
            return v.imag


def bdry_constraints():
    return [EM2Da_E]


class EM2Da_E(EM2Da_Bdry):
    has_essential = True
    vt = Vtable(data)

    def get_essential_idx(self, kfes):
        if kfes > 2:
            return
        return self._sel_index

    def apply_essential(self, engine, gf, real=False, kfes=0):
        if kfes > 1:
            return
        if real:
            dprint1("Apply Ess.(real)" + str(self._sel_index))
        else:
            dprint1("Apply Ess.(imag)" + str(self._sel_index))

        Erphiz = self.vt.make_value_or_expression(self)[0]

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns

        coeff1 = VCoeff(3, Erphiz, ind_vars, l, g, return_complex=True)

        mesh = engine.get_mesh(mm=self)
        ibdr = mesh.bdr_attributes.ToList()
        bdr_attr = [0]*mesh.bdr_attributes.Max()

        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        if kfes == 0:
            # coeff1 = Ep(2, Erphiz,
            #            self.get_root_phys().ind_vars,
            #            self._local_ns, self._global_ns,
            #            real = real)
            coeff1 = coeff1[[0, 2]]
            coeff1 = coeff1.get_realimag_coefficient(real)
            gf.ProjectBdrCoefficientTangent(coeff1,
                                            mfem.intArray(bdr_attr))
        elif kfes == 1:
            # coeff1 = rEt(Erphiz, self.get_root_phys().ind_vars,
            #            self._local_ns, self._global_ns,
            #            real = real)
            def r(ptx,):
                return ptx[0]
            r_coeff = func_to_numba_coeff_scalar(r,)

            coeff2 = coeff1[1]
            coeff3 = r_coeff*coeff2
            coeff3 = coeff3.get_realimag_coefficient(real)

            gf.ProjectBdrCoefficient(coeff3,
                                     mfem.intArray(bdr_attr))



'''
   Tangential H Boundary Condition

   (note)
    In MKSA,  1/mu curl E = -dH/dt = i omega H.
 
    Therefore, surface integral 
        \int W \dot (n \times 1/mu curl E) d\Omega
    becomes
        \int W \dot (n \times i omega H) d\Omega

    VectorFEBoundaryTangentLFIntegrator can be used.

   CopyRight (c) 2016-  S. Shiraiwa
''' 

import numpy as np

from petram.phys.phys_const import mu0, epsilon0

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model  import Phys, VectorPhysCoefficient
from petram.phys.em3d.em3d_base import EM3D_Bdry, EM3D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_H')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem
   
from petram.phys.vtable import VtableElement, Vtable      
data =  (('H', VtableElement('H', type='complex',
                             guilabel = 'H',
                             suffix =('x', 'y', 'z'),
                             default = [0,0,0],
                             tip = "magnetic field" )),)

class Ht(VectorPhysCoefficient):
   def __init__(self, *args, **kwargs):
       omega = kwargs.pop('omega', 1.0)
       self.fac = -1j*omega  # /mur
       super(Ht, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(Ht, self).EvalValue(x)
       nv = v*self.fac

       if self.real: return nv.real
       else: return nv.imag


def bdry_constraints():
   return [EM3D_H]
       
class EM3D_H(EM3D_Bdry):
    is_essential = False
    vt  = Vtable(data)
    
    def has_lf_contribution(self, kfes = 0):
        if kfes != 0: return False
        return True
    
    def add_lf_contribution(self, engine, b, real = True, kfes = 0):
        if kfes != 0: return
        if real:       
            dprint1("Add LF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + str(self._sel_index))
            
        freq, omega = self.get_root_phys().get_freq_omega()        

        h = self.vt.make_value_or_expression(self)
        coeff1 = Ht(3, h[0],  self.get_root_phys().ind_vars,
                           self._local_ns, self._global_ns,
                           real = real, omega = omega)
        self.add_integrator(engine, 'H', coeff1,
                            b.AddBoundaryIntegrator,
                            mfem.VectorFEBoundaryTangentLFIntegrator)



'''
   Vacuum region:
      However, can have arbitrary scalar epsilon_r, mu_r, sigma


'''
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.phys_model import PhysCoefficient
from petram.phys.coefficient import SCoeff, MCoeff
from petram.phys.em1d.em1d_base import EM1D_Bdry, EM1D_Domain
from petram.phys.em1d.em1d_vac import EM1D_Vac

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1D_Anisotropic')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('epsilonr', VtableElement('epsilonr', type='complex',
                                   guilabel='epsilonr',
                                   suffix=[('x', 'y', 'z'), ('x', 'y', 'z')],
                                   default=np.eye(3, 3),
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur', type='complex',
                              guilabel='mur',
                              default=1.0,
                              tip="relative permeability")),
        ('sigma', VtableElement('sigma', type='complex',
                                guilabel='sigma',
                                suffix=[('x', 'y', 'z'), ('x', 'y', 'z')],
                                default=np.zeros((3, 3)),
                                tip="contuctivity")),
        ('ky', VtableElement('ky', type='float',
                             guilabel='ky',
                             default=0.,
                             no_func=True,
                             tip="wave number in the y direction")),
        ('kz', VtableElement('kz', type='float',
                             guilabel='kz',
                             default=0.0,
                             no_func=True,
                             tip="wave number in the z direction")),)

'''
Expansion of matrix is as follows

                 [e_xx  e_xy  e_zz ][Ex]
                 [                 ][  ]
[Wx, Wy, Wz]  =  [e_yx  e_yy  e_yz ][Ey]
                 [                 ][  ]
                 [e_zx  e_zy  e_zz ][Ez]

'''

'''
class Epsilon(PhysCoefficient):
   #
   # - omega^2 * epsilon0 * epsilonr
   #
   def __init__(self, *args, **kwargs):
       self.omega = kwargs.pop('omega', 1.0)
       self.component = kwargs.pop('component', (0, 0))
       super(Epsilon, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(Epsilon, self).EvalValue(x)
       v = np.array(- v * epsilon0 * self.omega * self.omega, copy=False).reshape(3,3)
       v = v[self.component]
       if self.real:  return v.real
       else: return v.imag

def make_epsilon(*args, **kwargs):
    e = args[0]
    component = kwargs.get('component', None)
    omega = kwargs.get('omega', 1.0)
    real = kwargs.get('real', True)
    if any([isinstance(ee, str) for ee in e]):
        return Epsilon(*args, **kwargs)
    else:
        # conj is ignored..(this doesn't no meaning...)       
        if component is None:
            assert False, "index is not given"
            
        v = args[0][component]
        v = - v * epsilon0 * omega * omega        
        if real:  v = v.real
        else: v = v.imag
        return PhysConstant(float(v))
       
class Sigma(PhysCoefficient):
   #
   # -1j * omega * sigma
   #
   def __init__(self, *args, **kwargs):
       self.omega = kwargs.pop('omega', 1.0)
       self.component = kwargs.pop('component', (0, 0))       
       super(Sigma, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(Sigma, self).EvalValue(x)
       v = np.array(-1j * self.omega * v, copy=False).reshape(3,3)
       v = v[self.component]       
       if self.real:  return v.real
       else: return v.imag
       
def make_sigma(*args, **kwargs):
    e = args[0]
    component = kwargs.get('component', None)
    omega = kwargs.get('omega', 1.0)
    real = kwargs.get('real', True)
    if any([isinstance(ee, str) for ee in e]):
        return Sigma(*args, **kwargs)
    else:
        # conj is ignored..(this doesn't no meaning...)       
        if component is None:
            assert False, "index is not given"
            
        v = args[0][component]
        v = -1j * omega * v        
        if real:  v = v.real
        else: v = v.imag
        return PhysConstant(float(v))
'''


def Epsilon_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # - omega^2 * epsilon0 * epsilonr
    fac = -epsilon0 * omega * omega * cnorm
    return MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)


def Sigma_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # v = - 1j * self.omega * v
    fac = - 1j * omega*cnorm
    return MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)


def Mu_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # v = mu * v
    fac = mu0*cnorm
    # return MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return SCoeff([exprs], ind_vars, l, g, return_complex=True, scale=fac)


def domain_constraints():
    return [EM1D_Anisotropic]


class EM1D_Anisotropic(EM1D_Vac):
    vt = Vtable(data)
    #nlterms = ['epsilonr']

    def get_coeffs2(self, r, c):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s, ky, kz = self.vt.make_value_or_expression(self)

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns

        coeff1 = Epsilon_Coeff(e, ind_vars, l, g, omega, cnorm)
        coeff2 = Mu_Coeff(m, ind_vars, l, g, omega, cnorm)
        coeff3 = Sigma_Coeff(s, ind_vars, l, g, omega, cnorm)

        ec = coeff1[r, c]
        sc = coeff3[r, c]

        return ec, sc, ky, kz

    def add_bf_contribution(self, engine, a, real=True, kfes=0):

        ec, sc, ky, kz = self.get_coeffs2(kfes, kfes)
        self.set_integrator_realimag_mode(real)

        super(EM1D_Anisotropic, self).add_bf_contribution(engine, a, real=real,
                                                          kfes=kfes, ecsc=(ec, sc))

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real=True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))

        # adding mu. this part is the same as vacuum
        super(EM1D_Anisotropic, self).add_mix_contribution(engine, mbf, r, c, is_trans,
                                                           real=real)

        ec, sc, ky, kz = self.get_coeffs2(r, c)
        self.set_integrator_realimag_mode(real)

        self.add_integrator(engine, 'epsilonr', ec, mbf.AddDomainIntegrator,
                            mfem.MixedScalarMassIntegrator)
        self.add_integrator(engine, 'sigma', sc, mbf.AddDomainIntegrator,
                            mfem.MixedScalarMassIntegrator)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        e, m, s, ky, kz = self.vt.make_value_or_expression(self)

        if len(self._sel_index) == 0:
            return

        add_constant(v, 'ky', suffix, np.float64(ky),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        add_constant(v, 'kz', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        self.do_add_matrix_expr(v, suffix, ind_vars, 'epsilonr', e)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'smur', m, add_diag=3)
        self.do_add_matrix_expr(v, suffix, ind_vars, 'sigma', s)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')



from petram.mfem_config import use_parallel
import os
import numpy as np

from petram.solver.parametric_scanner import DefaultParametricScanner
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('PortScanner')
dprint0 = debug.regular_print('PortScanner', True)
format_memory_usage = debug.format_memory_usage

if use_parallel:
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    smyid = '.'+'{:0>6d}'.format(myid)
    from mfem.common.mpi_debug import nicePrint
else:
    myid = 0
    smyid = ''


class PortScanner(DefaultParametricScanner):
    ''' 
    Scanner for port BC amplitude

    PortScan([1,2,3], amplitude=1, phase=0.0, phys='first')
    PortScan([1,2,3], amplitude=1, phase=0.0, phys='EM3D1')
    PortScan(amplitude=1, phase=0.0)  " scan all port of first phys em model"

    # user need to load this in the global namelist

    '''

    def __init__(self, *args, **kwargs):
        self.amplitude = kwargs.pop("amplitude", 1)
        self.phase = kwargs.pop("phase", 0.0)
        self.phys_name = kwargs.pop("phys", 'first')
        self.smat_name = kwargs.pop("probe", 'Smat')

        if len(args) == 0:
            self._names = None
            self.port = 'all'
            data = []
        else:
            self.port = [str(x) for x in args[0]]
            data = list(range(len(self.port)))
            names = [str(x) for x in self.port]
            self._names = ["ports_"+"_".join(names)]

        self.smat = None
        DefaultParametricScanner.__init__(self, data=data)

    def get_probes(self):
        return [self.smat_name]

    def format_data(self, data):
        return "port "+self.port[data]

    def set_data_from_model(self, root):
        from petram.phys.em3d.em3d_port import EM3D_Port
        from petram.phys.em3d.em3d_portarray import EM3D_PortArray
        #from petram.phys.em2d.em2d_port import EM2D_Port
        from petram.phys.em2da.em2da_port import EM2Da_Port
        from petram.phys.em1d.em1d_port import EM1D_Port

        phys_root = root['Phys']

        if self._names is None:  # automatic scan of all ports in first EM model
            iports = []
            for phys in phys_root:
                if (self.phys_name != 'first' and
                        not phys != self.phys_name):
                    continue
                found = False
                for obj in phys_root[phys].walk():
                    if not obj.enabled:
                        continue
                    if (isinstance(obj, EM3D_Port) or
                        isinstance(obj, EM2Da_Port) or
                            isinstance(obj, EM1D_Port)):

                        iports.append(str(obj.port_idx))
                        found = True
                    elif isinstance(obj, EM3D_PortArray):
                        for i in range(len(obj.sel_index)):
                            iports.append(str(obj.port_idx) + '_' + str(i+1))
                        found = True
                if found:
                    self.phys_name = phys
                    dprint1("Scanned all ports in " + phys)
                    break
            self.port = iports
            names = ["port."+str(x) for x in self.port]
            data = list(range(len(self.port)))

            self.set_data(data)

        elif self.phys_name == 'first':  # set phys_name to first EM model
            for phys in phys_root:
                found = False
                for obj in phys_root[phys].walk():
                    if not obj.enabled:
                        continue
                    if (isinstance(obj, EM3D_Port) or
                        isinstance(obj, EM2Da_Port) or
                        isinstance(obj, EM1D_Port) or
                            isinstance(obj, EM3D_PortArray)):
                        found = True
                if found:
                    self.phys_name = phys
                    break
        dprint1("Port scanner is used for :" + self.phys_name)

    def apply_param(self, data):
        from petram.phys.em3d.em3d_port import EM3D_Port
        from petram.phys.em3d.em3d_portarray import EM3D_PortArray
        #from petram.phys.em2d.em2d_port import EM2D_Port
        from petram.phys.em2da.em2da_port import EM2Da_Port
        from petram.phys.em1d.em1d_port import EM1D_Port

        dprint1("Port Scanner: Port Amplitude: " + str(data))

        for phys in self.target_phys:
            if phys.name() != self.phys_name:
                continue
            for obj in phys.walk():
                if not obj.enabled:
                    continue

                if (isinstance(obj, EM3D_Port) or
                    isinstance(obj, EM2Da_Port) or
                        isinstance(obj, EM1D_Port)):

                    obj.inc_amp_txt = '0.0'
                    if self.port[data] == str(obj.port_idx):
                        obj.inc_amp_txt = str(self.amplitude)
                        obj.inc_phase_txt = str(self.phase)
                        dprint1("Port Scanner: Setting port:",
                                obj, "=", obj.inc_amp_txt)

                elif isinstance(obj, EM3D_PortArray):
                    obj.inc_phase_txt = str(self.phase)
                    dd = [0.0]*len(obj.sel_index)
                    for i in range(len(obj.sel_index)):
                        n = str(obj.port_idx) + '_' + str(i+1)
                        if self.port[data] == n:
                            dd[i] = self.amplitude
                    obj.inc_amp_txt = ','.join([str(x) for x in dd])
                    dprint1("Port Scanner: Setting port:",
                            obj, "=", obj.inc_amp_txt)

    @property
    def names(self):
        '''
        suposed to return parameternames
        '''
        return self._names

    def collect_probe_signals(self, engine, dirs):
        from petram.phys.em3d.em3d_port import EM3D_Port
        from petram.phys.em3d.em3d_portarray import EM3D_PortArray
        #from petram.phys.em2d.em2d_port import EM2D_Port
        from petram.phys.em2da.em2da_port import EM2Da_Port
        from petram.phys.em1d.em1d_port import EM1D_Port

        pnames = []
        for phys in self.target_phys:
            if phys.name() != self.phys_name:
                continue

            file_suffix = phys.dep_vars_suffix
            for obj in phys.walk():
                if (isinstance(obj, EM3D_Port) or
                    isinstance(obj, EM2Da_Port) or
                        isinstance(obj, EM1D_Port)):
                    # this assumes port BC define only 1 probe signal, which
                    # is true at this moment
                    pnames.append((int(obj.port_idx), obj.get_probes()[0]))

                elif isinstance(obj, EM3D_PortArray):
                    dprint1("Smatrix collection is not supported for PortArray")

        ports = [x[0] for x in sorted(pnames)]
        pnames = [x[1] for x in sorted(pnames)]

        from petram.sol.probe import list_probes, load_probe,  Probe

        od = os.getcwd()
        filenames, probenames = list_probes(od)

        files = dict(zip(probenames, filenames))
        smat = []
        for p in pnames:
            if p+smyid not in files:
                # myid != 0 does not have data.
                continue
            f = files[p+smyid]
            xdata, ydata = load_probe(f)
            smat.append(ydata)

        if len(smat) != 0:
            p = Probe(self.smat_name+file_suffix, xnames=["ports"])
            for idx, item in zip(ports, smat):
                p.append_value(item, t=idx)
            p.write_file()

            # store Smatrix to _variables
            smat = np.vstack(smat)
            from petram.helper.variables import Constant
            engine.model._variables[self.smat_name +
                                    file_suffix] = Constant(smat)

        if use_parallel:
            from petram.helper.mpi_recipes import scatter_vector
            if myid == 0:
                s = np.array(smat.shape, dtype=np.int32)
                tmp = smat.flatten().astype(np.complex128)
            else:
                s = None
                tmp = None

            s = MPI.COMM_WORLD.bcast(s)
            tmp = MPI.COMM_WORLD.bcast(tmp)
            smat = tmp.reshape(s)

        # store this so that we can access this from sanner in global_ns
        self.Smat = smat


PortScan = PortScanner



'''
   local-K plasma.
'''
from petram.phys.common.rf_dispersion_lkplasma import (vtable_data0,
                                                       default_kpe_option,
                                                       kpe_options,
                                                       term_options,
                                                       default_term_option)

from petram.phys.common.rf_dispersion_coldplasma import (col_model_options,
                                                         default_col_model,)

import numpy as np

from petram.mfem_config import use_parallel, get_numba_debug

from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.numba_coefficient import NumbaCoefficient

from petram.phys.phys_model import MatrixPhysCoefficient, PhysCoefficient, PhysConstant, PhysMatrixConstant
from petram.phys.em1d.em1d_base import EM1D_Bdry, EM1D_Domain
from petram.phys.em1d.em1d_vac import EM1D_Vac

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1D_LocalKPlasma')

if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
else:
    import mfem.ser as mfem
    myid = 0

vtable_data = vtable_data0.copy()
vtable_data.extend(
    [('ky', VtableElement('ky', type='float',
                          guilabel='ky',
                          default=0.,
                          no_func=True,
                          tip="wave number in the y direction")),
     ('kz', VtableElement('kz', type='float',
                          guilabel='kz',
                          default=0.0,
                          no_func=True,
                          tip="wave number in the z direction"))])

kpe_alg_options = ["std", "em1d"]


def domain_constraints():
    return [EM1D_LocalKPlasma]


class EM1D_LocalKPlasma(EM1D_Vac):
    allow_custom_intorder = False
    vt = Vtable(vtable_data)

    def __init__(self, **kargs):
        super(EM1D_LocalKPlasma, self).__init__(**kargs)
        self._jited_coeff = None

    @classmethod
    def fancy_menu_name(cls):
        return "HotPlasma(local-k)"

    @classmethod
    def fancy_tree_name(cls):
        return "LocalKPlasma"

    def get_possible_child(self):
        return []

    def attribute_set(self, v):
        EM1D_Vac.attribute_set(self, v)
        v["kpe_mode"] = default_kpe_option
        v["kpe_alg"] = kpe_alg_options[0]
        v["col_model"] = default_col_model
        v["lk_terms"] = None
        return v

    def config_terms(self, evt):
        from petram.phys.common.rf_lk_term_panel import ask_rf_lk_terms

        self.vt.preprocess_params(self)
        ret = self.vt.make_value_or_expression(self)
        charges = ret[7]

        num_ions = len(charges)
        win = evt.GetEventObject()
        value = ask_rf_lk_terms(win, num_ions, self.lk_terms)
        self.lk_terms = value

    def panel1_param(self):
        panels = super(EM1D_LocalKPlasma, self).panel1_param()
        panels.append(["kpe mode", None, 1, {"values": kpe_options}])
        panels.append(["kpe alg.", None, 1, {"values": kpe_alg_options}])
        panels.append(["col. model", None, 1, {"values": col_model_options}])
        panels.extend([  # ["contributions", "", 2, None],
            [None, None, 341, {"label": "Customize terms",
                               "func": "config_terms",
                               "sendevent": True,
                               "noexpand": True}], ])
        return panels

    def get_panel1_value(self):
        values = super(EM1D_LocalKPlasma, self).get_panel1_value()
        values.extend([self.kpe_mode, self.kpe_alg, self.col_model, self])
        return values

    def import_panel1_value(self, v):
        check = super(EM1D_LocalKPlasma, self).import_panel1_value(v[:-4])
        self.kpe_mode = v[-4]
        self.kpe_alg = v[-3]
        self.col_model = v[-2]
        return check

    @property
    def jited_coeff(self):
        return self._jited_coeff

    def compile_coeffs(self):
        self._jited_coeff = self.get_coeffs()

    def get_coeffs(self):
        _freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        B, dens_e, t_e, dens_i, t_i, t_c, masses, charges, kpakpe, kpevec, ky, kz = self.vt.make_value_or_expression(
            self)
        ind_vars = self.get_root_phys().ind_vars
        kpe_mode = self.kpe_mode
        kpe_alg = self.kpe_alg

        from petram.phys.common.rf_dispersion_lkplasma import value2flags

        num_ions = len(charges)
        terms = value2flags(num_ions, self.lk_terms)

        from petram.phys.common.rf_dispersion_lkplasma import build_coefficients

        coeff1, coeff2, coeff3 = build_coefficients(ind_vars, omega, B, t_c, dens_e, t_e,
                                                    dens_i, t_i, masses, charges, kpakpe, kpevec,
                                                    kpe_mode, self.col_model, cnorm,
                                                    self._global_ns, self._local_ns,
                                                    kpe_alg=kpe_alg, sdim=1, kymode=ky, kzmode=kz, terms=terms)
        return coeff1, coeff2, coeff3, ky, kz

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        if real:
            dprint1("Add BF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add BF contribution(imag)" + str(self._sel_index))

        coeff1, coeff2, coeff3, ky, kz = self.jited_coeff
        self.set_integrator_realimag_mode(real)

        # if self.has_pml():
        #    coeff1 = self.make_PML_coeff(coeff1)
        #    coeff2 = self.make_PML_coeff(coeff2)
        #    coeff3 = self.make_PML_coeff(coeff3)

        ec = coeff1[kfes, kfes]
        sc = coeff3[kfes, kfes]

        self.add_integrator(engine, 'epsilonr', ec, a.AddDomainIntegrator,
                            mfem.MassIntegrator)
        self.add_integrator(engine, 'sigma', sc, a.AddDomainIntegrator,
                            mfem.MassIntegrator)

        coeff4 = 1./coeff2[0, 0]
        #
        #
        #
        if kfes == 0:  # Ex
            imu = coeff4*(ky**2 + kz**2)
            self.add_integrator(engine, 'mur', imu, a.AddDomainIntegrator,
                                mfem.MassIntegrator)

        elif kfes == 1 or kfes == 2:  # Ey and Ez
            imu = coeff4
            self.add_integrator(engine, 'mur1', imu, a.AddDomainIntegrator,
                                mfem.DiffusionIntegrator)

            if kfes == 1:
                fac = kz*kz
            if kfes == 2:
                fac = ky*ky

            imu = coeff4*fac

            self.add_integrator(engine, 'mur2', imu, a.AddDomainIntegrator,
                                mfem.MassIntegrator)

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real=True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))

        coeff1, coeff2, coeff3, ky, kz = self.jited_coeff
        self.set_integrator_realimag_mode(real)

        # super(EM1D_LocalKPlasma, self).add_mix_contribution(engine, mbf, r, c, is_trans,
        #                                                   real = real)
        ec = coeff1[r, c]
        sc = coeff3[r, c]

        self.add_integrator(engine, 'epsilonr', ec, mbf.AddDomainIntegrator,
                            mfem.MixedScalarMassIntegrator)
        self.add_integrator(engine, 'sigma', sc, mbf.AddDomainIntegrator,
                            mfem.MixedScalarMassIntegrator)

        coeff4 = 1./coeff2[0, 0]
        if r == 0 and c == 1:
            imu = coeff4*(1j*ky)
            itg = mfem.MixedScalarDerivativeIntegrator
        elif r == 0 and c == 2:
            imu = coeff4*(1j*kz)
            itg = mfem.MixedScalarDerivativeIntegrator
        elif r == 1 and c == 0:
            imu = coeff4*(1j*ky)
            itg = mfem.MixedScalarWeakDerivativeIntegrator
        elif r == 1 and c == 2:
            imu = coeff4*(-ky*kz)
            itg = mfem.MixedScalarMassIntegrator
        elif r == 2 and c == 0:
            imu = coeff4*(1j*kz)
            itg = mfem.MixedScalarWeakDerivativeIntegrator
        elif r == 2 and c == 1:
            imu = coeff4*(-ky*kz)
            itg = mfem.MixedScalarMassIntegrator
        else:
            assert False, "Something is wrong..if it comes here;D"

        self.add_integrator(engine, 'mur', imu,
                            mbf.AddDomainIntegrator, itg)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        if len(self._sel_index) == 0:
            return

        freq, omega = self.get_root_phys().get_freq_omega()
        B, dens_e, t_e, dens_i, t_i, t_c, masses, charges, kpakpe, kpevec, ky, kz = self.vt.make_value_or_expression(
            self)

        kpe_mode = self.kpe_mode
        kpe_alg = self.kpe_alg

        add_constant(v, 'ky', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)
        add_constant(v, 'kz', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        from petram.phys.common.rf_dispersion_lkplasma import value2flags

        num_ions = len(charges)
        terms = value2flags(num_ions, self.lk_terms)

        from petram.phys.common.rf_dispersion_lkplasma import build_variables

        ss = self.parent.parent.name()+'_'+self.name()  # phys module name + name
        ret = build_variables(v, ss, ind_vars,
                              omega, B, t_c, dens_e, t_e,
                              dens_i, t_i, masses, charges,
                              kpakpe, kpevec, kpe_mode, kpe_alg, self.col_model,
                              self._global_ns, self._local_ns,
                              sdim=1, terms=terms)

        from petram.phys.common.rf_dispersion_lkplasma import add_domain_variables_common
        add_domain_variables_common(self, ret, v, suffix, ind_vars)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')

        return



from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain

from petram.phys.vtable import VtableElement, Vtable   
data =  (('label1', VtableElement(None, 
                                  guilabel = 'Perfect Magnetic Conductor',
                                  default =   "Ht = 0",
                                  tip = "this is a natural BC" )),)
def bdry_constraints():
   return [EM2D_PMC]

class EM2D_PMC(EM2D_Bdry):
    is_essential = False
    nlterms = []
    vt  = Vtable(data)          
    



'''
   Vacuum region:
      However, can have arbitrary scalar epsilon_r, mu_r, sigma


'''
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.coefficient import SCoeff
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.phys_model import PhysCoefficient, PhysConstant
from petram.phys.em3d.em3d_base import EM3D_Bdry, EM3D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_Vac')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('epsilonr', VtableElement('epsilonr', type='complex',
                                   guilabel='epsilonr',
                                   default=1.0,
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur', type='complex',
                              guilabel='mur',
                              default=1.0,
                              tip="relative permeability")),
        ('sigma', VtableElement('sigma', type='complex',
                                guilabel='sigma',
                                default=0.0,
                                tip="contuctivity")),)


def Epsilon_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # - omega^2 * epsilon0 * epsilonr
    fac = -epsilon0 * omega * omega * cnorm
    coeff = SCoeff(exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def Sigma_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # v = - 1j * self.omega * v
    fac = - 1j * omega * cnorm
    coeff = SCoeff(exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def Mu_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # v = mu * v
    fac = mu0/cnorm
    coeff = SCoeff(exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def domain_constraints():
    return [EM3D_Vac]


class EM3D_Vac(EM3D_Domain):
    vt = Vtable(data)
    #nlterms = ['epsilonr']

    def get_possible_child(self):
        from .em3d_pml import EM3D_LinearPML
        return [EM3D_LinearPML]

    def has_bf_contribution(self, kfes):
        if kfes == 0:
            return True
        else:
            return False

    def get_coeffs(self):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s = self.vt.make_value_or_expression(self)

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns
        coeff1 = Epsilon_Coeff([e], ind_vars, l, g, omega, cnorm)
        coeff2 = Mu_Coeff([m], ind_vars, l, g, omega, cnorm)
        coeff3 = Sigma_Coeff([s], ind_vars, l, g, omega, cnorm)

        dprint1("epsr, mur, sigma " + str(coeff1) +
                " " + str(coeff2) + " " + str(coeff3))

        return coeff1, coeff2, coeff3

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        if kfes != 0:
            return
        if real:
            dprint1("Add BF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add BF contribution(imag)" + str(self._sel_index))

        # e, m, s
        coeff1, coeff2, coeff3 = self.get_coeffs()
        self.set_integrator_realimag_mode(real)

        if self.has_pml():
            coeff1 = self.make_PML_coeff(coeff1)
            coeff2 = self.make_PML_coeff(coeff2)
            coeff3 = self.make_PML_coeff(coeff3)
            coeff2 = coeff2.inv()
        else:
            coeff2 = coeff2**(-1)

        self.add_integrator(engine, 'epsilonr', coeff1,
                            a.AddDomainIntegrator,
                            mfem.VectorFEMassIntegrator)
        self.add_integrator(engine, 'mur', coeff2,
                            a.AddDomainIntegrator,
                            mfem.CurlCurlIntegrator)
        self.add_integrator(engine, 'sigma', coeff3,
                            a.AddDomainIntegrator,
                            mfem.VectorFEMassIntegrator)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        if len(self._sel_index) == 0:
            return

        e, m, s = self.vt.make_value_or_expression(self)

        self.do_add_scalar_expr(v, suffix, ind_vars,
                                'sepsilonr', e, add_diag=3)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'smur', m, add_diag=3)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'ssigma', s, add_diag=3)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')



'''
  export_noralvalues.py

  a helper routine to export nordalvalues from  
  a pair of solution and mesh file.

  perform 3D interpolation of using GridData

  sample usage:

     #  go to Petra-M solution directory

     >>>  import petram.helper.export_nodalvalues as exporter
     >>>  solset = exporter.load_sol(fesvar = 'E')
     >>>  values = exporter.get_nodalvalues(solset)

     >>>  R =   np.arange(0.1, 1.5, 100)
     >>>  Z =   np.arange(-1.,  1., 100)
     >>>  Phi = np.arange(0, 2*np.pi., 100)

  sample script

     x = np.linspace(0, 0.007, 5)
     y = np.linspace(0, 0.2, 50)
     z = np.linspace(0, 0.06, 10)
     X, Y, Z = np.meshgrid(x, y, z)

     #save E
     exporter.export_interpolated_data(path, X, Y, Z, 'E',
                          # vdim = 3, complex = True,
                          nproc = 1, ncfile = 'data.nc')
     #save E and B
     exporter.export_interpolated_data2(path, X, Y, Z, 
                                        1e8, 'E',
                                        nproc = 1, ncfile = 'data.nc')




  Author: S. Shiraiwa

'''
import os
import six
import numpy as np
import time
from time import gmtime, strftime

import mfem.ser as mfem
import petram

def load_sol(solfiles, fesvar, refine = 0, ):
    '''
    read sol files in directory path. 
    it reads only a file for the certain FES variable given by fesvar
    '''
    from petram.sol.solsets import find_solfiles, MeshDict
#    solfiles = find_solfiles(path)

    def fname2idx(t):    
           i = int(os.path.basename(t).split('.')[0].split('_')[-1])
           return i
    solfiles = solfiles.set
    solset = []

    for meshes, solf, in solfiles:
        s = {}
        for key in six.iterkeys(solf):
            name = key.split('_')[0]
            if name != fesvar: continue
            idx_to_read = int(key.split('_')[1])
            break

    for meshes, solf, in solfiles:
        idx = [fname2idx(x) for x in meshes]
        meshes = {i:  mfem.Mesh(str(x), 1, refine) for i, x in zip(idx, meshes) if i == idx_to_read}
        meshes=MeshDict(meshes) # to make dict weakref-able
        ### what is this refine = 0 !?
        for i in meshes.keys():
            meshes[i].ReorientTetMesh()
            meshes[i]._emesh_idx = i

        s = {}
        for key in six.iterkeys(solf):
            name = key.split('_')[0]
            if name != fesvar: continue

            fr, fi =  solf[key]
            i = fname2idx(fr)
            m = meshes[i]
            solr = (mfem.GridFunction(m, str(fr)) if fr is not None else None)
            soli = (mfem.GridFunction(m, str(fi)) if fi is not None else None)
            if solr is not None: solr._emesh_idx = i
            if soli is not None: soli._emesh_idx = i
            s[name] = (solr, soli)
        solset.append((meshes, s))
        
    return solset

def get_nodalvalues(solset,
                    curl = False,
                    grad = False,
                    div  = False):

    if grad:
        assert False, "evaluating Grad is not implemented"
    if div:
        assert False, "evaluating Div is not implemented"

    import petram.helper.eval_deriv as eval_deriv

    from collections import defaultdict
    nodalvalues =defaultdict(list)

    for meshes, s in solset:   # this is MPI rank loop
        for name in s:
            gfr, gfi = s[name]
            m = gfr.FESpace().GetMesh()
            size = m.GetNV()

            ptx = np.vstack([m.GetVertexArray(i) for i in range(size)])
            
            gfro, gfio = gfr, gfi
            if curl:
                gfr, gfi, extra = eval_deriv.eval_curl(gfr, gfi)            
            dim = gfr.VectorDim()
            
            ret = np.zeros((size, dim), dtype = float)            
            for comp in range(dim):
                values = mfem.Vector()
                gfr.GetNodalValues(values, comp+1)
                ret[:, comp] = values.GetDataArray()
                values.StealData()

            if gfi is None:
                nodalvalues[name].append((ptx, ret, gfr))
                continue

            ret2 = np.zeros((size, dim), dtype = float)                        
            for comp in range(dim):
                values = mfem.Vector()
                gfi.GetNodalValues(values, comp+1)

                if ret2 is None:
                    ret2 = np.zeros((values.Size(), dim), dtype = float)

                ret2[:, comp] = values.GetDataArray()
                values.StealData()

            ret = ret + 1j*ret2
            nodalvalues[name].append((ptx, ret, gfro))

            
    nodalvalues.default_factory = None
    return nodalvalues

def make_mask(values, X, Y, Z, mask_start = 0, logfile = None):
    '''
    mask for interpolation
    '''
    mask= np.zeros(len(X.flatten()), dtype=int) - 1

    for kk, data in enumerate(values):
       ptx, ret, gfr = data

       xmax = np.max(ptx[:, 0]);xmin = np.min(ptx[:, 0])
       ymax = np.max(ptx[:, 1]);ymin = np.min(ptx[:, 1])
       zmax = np.max(ptx[:, 2]);zmin = np.min(ptx[:, 2])

       i1 = np.logical_and(xmax >= X.flatten(), xmin <= X.flatten())
       i2 = np.logical_and(ymax >= Y.flatten(), ymin <= Y.flatten())
       i3 = np.logical_and(zmax >= Z.flatten(), zmin <= Z.flatten())
       ii = np.logical_and(np.logical_and(i1, i2), i3)

       mesh = gfr.FESpace().GetMesh()

       XX = X.flatten()[ii]
       YY = Y.flatten()[ii]
       ZZ = Z.flatten()[ii]
       size = len(XX)

       m = mfem.DenseMatrix(3, size)       
       ptx = np.vstack([XX, YY, ZZ])

       if logfile is not None:
          txt = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
          logfile.write(txt + "\n")
          logfile.write("calling FindPoints : size = "  + str(XX.shape) + " " +
                        str(kk+1) + '/' + str(len(values)) + "\n")
       
       m.Assign(ptx)
       ips = mfem.IntegrationPointArray()
       elem_ids = mfem.intArray()
       
       pts_found = mesh.FindPoints(m, elem_ids, ips)

       ii = np.where(ii)[0][np.array(elem_ids.ToList()) !=-1 ]
       mask[ii] = kk + mask_start
       if logfile is not None:
           logfile.write("done\n")
            
    return mask


def interp3D(values, X, Y, Z, mask, vdim=1, complex = False, mask_start=0,
             logfile = None):
    from scipy.interpolate import griddata
    
    X = X.flatten()
    Y = Y.flatten()
    Z = Z.flatten()        
    
    size = len(X.flatten())
    if complex:
        res = np.zeros((vdim, size), dtype=np.complex128)
    else:
        res = np.zeros((vdim, size), dtype=np.float64)
    
    for kk, data in enumerate(values):

        if logfile is not None:
            txt = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
            logfile.write(txt + "\n")
            logfile.write("processing interp3D :"  + str(kk+1) + '/' + str(len(values))+ "\n")
        
        idx = mask == kk + mask_start
        ptx, ret, gfr = data

        for ii in range(vdim):
           print(ptx.shape, ret.shape)
           res[ii, idx] = griddata(ptx, ret[:, ii].flatten(),
                                   (X[idx], Y[idx], Z[idx]))
           
        if logfile is not None:
            logfile.write("done\n")

    if vdim == 1: res = res.flatten()       
    return res

import multiprocessing as mp

class exporter_child(mp.Process):
    def __init__(self, result_queue, myid, rank, *args, **kwargs):

        mp.Process.__init__(self)
        self.result_queue = result_queue
        self.myid = myid
        self.rank = rank
        self.logfile = open('export.log.'+str(myid), 'w', buffering=0)
        self.args = args
        self.call_curl = kwargs.pop('curl', False)
        
 
    def data_partition(self, m, num_proc, myid):
        min_nrows  = m / num_proc
        extra_rows = m % num_proc
        start_row  = min_nrows * myid + (extra_rows if extra_rows < myid else myid)
        end_row    = start_row + min_nrows + (1 if extra_rows > myid else 0)
        nrows   = end_row - start_row
        return start_row, end_row
        
    def run(self):
        try:
            path, X, Y, Z, fesvar, vdim, complex = self.args
        
            from petram.sol.solsets import find_solfiles, MeshDict
            solfiles = find_solfiles(path)
            st, et = self.data_partition(len(solfiles.set), self.rank, self.myid)
            self.logfile.write('partition '+str(st)+':'+str(et)+'\n')
         
            solset = load_sol(solfiles[st:et], fesvar)
            self.logfile.write('calling nodalvalues\n')
            values = get_nodalvalues(solset, curl = self.call_curl)
            self.logfile.write('calling make_mask\n')            
            mask   = make_mask(values[fesvar], X, Y, Z, mask_start = st,
                               logfile = self.logfile)
            self.logfile.write('calling interp3D\n')                        
            data = interp3D(values[fesvar], X, Y, Z, mask, vdim=vdim,
                            complex=complex, mask_start = st, logfile = self.logfile)
            
            idx = np.where(mask != -1)[0]
            if vdim == 1:
                data = data[idx]
            else:
                data = data[:, idx]                
            #self.result_queue.put((mask, data))
            self.result_queue.put((idx, mask[idx],  data))
        except:
            import traceback
            self.result_queue.put((None, None, traceback.format_exc()))
        self.logfile.close()

        
def export_interpolated_data(path, X, Y, Z, fesvar,
                             vdim = 1, complex = False, nproc = 1,
                             ncfile = 'data.nc',
                             curl = False, return_mask = False):
    
    from netCDF4 import Dataset
    
    results= mp.JoinableQueue() 
    workers = [None]*nproc
    
    for i in range(nproc):
         w = exporter_child(results, i, nproc,
                            path, X, Y, Z, fesvar, vdim, complex,
                            curl=curl)
         workers[i] = w
         time.sleep(0.1)
    for w in workers:
         w.daemon = True
         w.start()
         
    res = [results.get() for x in range(len(workers))]
    for x in range(len(workers)):
         results.task_done()
         
    size = len(X.flatten())
    if complex:
        ans = np.zeros((vdim, size), dtype=np.complex128)
    else:
        ans = np.zeros((vdim, size), dtype=np.float64)
        
    mask= np.zeros(len(X.flatten()), dtype=int) - 1
    for idx, mm, dd in res:
        if mm is None:
            print(dd)
            assert False, "Child Process Failed"
        else:
            if idx.size == 0: continue
            print("here", idx.shape, dd.shape)
            if vdim == 1:
                ans[idx] = dd
            else:
                ans[:,idx] = dd
            mask[idx] = mm

    ans = ans.reshape(-1, X.shape[0], X.shape[1], X.shape[2])
    mask = mask.reshape(X.shape[0], X.shape[1], X.shape[2])    

    if ncfile != '':
        nc = Dataset(ncfile, "w", format='NETCDF4')
        nc.createDimension('vdim', vdim)
        nc.createDimension('dim_0', X.shape[0])
        nc.createDimension('dim_1', X.shape[1])
        nc.createDimension('dim_2', X.shape[2])
        if complex:        
            a_real = nc.createVariable(fesvar+'_real', np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_real[:] = ans.real

            a_imag = nc.createVariable(fesvar+'_imag', np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_imag[:] = ans.imag
        else:
            a_real = nc.createVariable(fesvar, np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_real[:] = ans
            
        xx = nc.createVariable('X', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))
        yy = nc.createVariable('Y', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))
        zz = nc.createVariable('Z', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))
        rank = nc.createVariable('rank', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))

        xx[:] = X
        yy[:] = Y
        zz[:] = Z
        rank[:] = mask

        nc.close()

    if return_mask:
        return ans, mask
    else:
        return ans


def export_interpolated_data2(path, X, Y, Z, freq, fesvar='E',
                              vdim = 3, complex = True,                              
                              nproc = 8,  ncfile= 'EBdata.nc'):

    from netCDF4 import Dataset
    
    Edata, mask = export_interpolated_data(path, X, Y, Z, fesvar,
                                           vdim = vdim, complex = True,
                                           nproc = nproc, return_mask = True)
    Bdata, mask = export_interpolated_data(path, X, Y, Z, fesvar,
                                           vdim = vdim, complex = True, curl=True,
                                           nproc = nproc, return_mask = True)
    

    fesvar = ''.join([x for x in fesvar if not x.isdigit()])
    omega = 2*np.pi*freq
    if ncfile != '':
        nc = Dataset(ncfile, "w", format='NETCDF4')
        nc.createDimension('vdim', vdim)
        nc.createDimension('dim_0', X.shape[0])
        nc.createDimension('dim_1', X.shape[1])
        nc.createDimension('dim_2', X.shape[2])
        if complex:        
            a_real = nc.createVariable(fesvar+'_real', np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_real[:] = Edata.real

            a_imag = nc.createVariable(fesvar+'_imag', np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_imag[:] = Edata.imag
            
            a_real = nc.createVariable('B_real', np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_real[:] = (Bdata/1j).real/1/omega

            a_imag = nc.createVariable('B_imag', np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_imag[:] = (Bdata/1j).imag/omega
            
        else:
            a_real = nc.createVariable(fesvar, np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_real[:] = Edata
            a_real = nc.createVariable('B', np.dtype('double'),
                                   ('vdim', 'dim_0', 'dim_1', 'dim_2'))
            a_real[:] = Bdata/omega
            
        xx = nc.createVariable('X', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))
        yy = nc.createVariable('Y', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))
        zz = nc.createVariable('Z', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))
        rank = nc.createVariable('rank', np.dtype('double'),
                               ('dim_0', 'dim_1', 'dim_2'))

        xx[:] = X
        yy[:] = Y
        zz[:] = Z
        rank[:] = mask

        nc.close()

'''
sample usage

x = np.linspace(0, 0.007, 5)
y = np.linspace(0, 0.2, 50)
z = np.linspace(0, 0.06, 10)
X, Y, Z = np.meshgrid(x, y, z)
data = exporter.export_interpolated_data(path, X, Y, Z, 'E',
                          vdim = 3, complex = True,
                          nproc = 1, ncfile = 'data.nc')

'''



              



'''
   Anistropic media:
      However, can have arbitrary scalar epsilon_r, mu_r, sigma

    Expansion of matrix is as follows

               [e_xy  e_12 ][Exy ]
[Wxy, Wz]   =  [           ][    ] = Wrz e_xy Erz + Wxy e_12 Ez
               [e_21  e_zz][Ez  ]

                                   + Wz e_21 Exy + Wz*e_zz*Exy


  Exy = Ex e_x + Ey e_y
  Ez =  Ez e_z

'''
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_Anisotropic')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

'''
from petram.phys.coefficient import PyComplexMatrixInvCoefficient as ComplexMatrixInv
from petram.phys.coefficient import PyComplexMatrixProductCoefficient as ComplexMatrixProduct
from petram.phys.coefficient import PyComplexMatrixSumCoefficient as ComplexMatrixSum
from petram.phys.coefficient import PyComplexMatrixSliceCoefficient as ComplexMatrixSlice
from petram.phys.coefficient import PyComplexMatrixAdjCoefficient as ComplexMatrixAdj
'''
from petram.phys.vtable import VtableElement, Vtable   
data =  (('epsilonr', VtableElement('epsilonr', type='complex',
                                     guilabel = 'epsilonr',
                                     suffix =[('x', 'y', 'z'), ('x', 'y', 'z')],
                                     default = np.eye(3),
                                     tip = "relative permittivity" )),
         ('mur', VtableElement('mur', type='complex',
                                     guilabel = 'mur',
                                     suffix =[('x', 'y', 'z'), ('x', 'y', 'z')],
                                     default = np.eye(3),
                                     tip = "relative permeability" )),
         ('sigma', VtableElement('sigma', type='complex',
                                     guilabel = 'sigma',
                                     suffix =[('x', 'y', 'z'), ('x', 'y', 'z')],
                                     default = np.zeros((3, 3)),
                                     tip = "contuctivity" )),
         ('kz', VtableElement('kz', type='float',
                                     guilabel = 'kz',
                                     default = 0.0,
                                     no_func = True,
                                     tip = "out-of-plane wave number" )),)

from petram.phys.coefficient import MCoeff
from petram.phys.phys_const import mu0, epsilon0

def Epsilon_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # - omega^2 * epsilon0 * epsilonr
    fac = -epsilon0 * omega * omega * cnorm
    return MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)

def Sigma_Coeff(exprs, ind_vars, l, g, omega,  cnorm):
    # v = - 1j * self.omega * v
    fac = - 1j * omega * cnorm
    return MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)

def Mu_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # v = mu * v
    fac = mu0/cnorm
    return MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)

from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain, EM2D_Domain_helper

def domain_constraints():
   return [EM2D_Anisotropic]
class EM2D_Anisotropic(EM2D_Domain, EM2D_Domain_helper):
    vt  = Vtable(data)
    #nlterms = ['epsilonr']
    def get_possible_child(self):
        from .em2d_pml      import EM2D_LinearPML
        return [EM2D_LinearPML]
    
    def has_bf_contribution(self, kfes):
        if kfes == 0: return True
        elif kfes == 1: return True        
        else: return False
        
    def has_mixed_contribution(self):
        return True

    def get_mixedbf_loc(self):
        '''
        r, c, and flag1, flag2 of MixedBilinearForm
           flag1 : take transpose
           flag2 : take conj
        '''
        return [(0, 1, 1, 1), (1, 0, 1, 1),]#(0, 1, -1, 1)]

    def get_coeffs(self):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s, kz = self.vt.make_value_or_expression(self)

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns
        coeff1 = Epsilon_Coeff(e, ind_vars, l, g, omega, cnorm)
        coeff2 = Mu_Coeff(m, ind_vars, l, g, omega, cnorm)
        coeff3 = Sigma_Coeff(s, ind_vars, l, g, omega, cnorm)

        #dprint1("epsr, mur, sigma " + str(coeff1) + " " + str(coeff2) + " " + str(coeff3))
        return coeff1, coeff2, coeff3, kz

    def get_coeffs_2(self):
        # e, m, s
        coeff1, coeff2, coeff3, kz = self.get_coeffs()
        '''
        coeff4 = ComplexMatrixSum(coeff1, coeff3)      # -> coeff4 = coeff1 + coeff3
        '''
        coeff4 = coeff1 + coeff3

        if self.has_pml():
            coeff4 = self.make_PML_coeff(coeff4)
            coeff2 = self.make_PML_coeff(coeff2)

        '''
        eps11 = ComplexMatrixSlice(coeff4, [0,1], [0,1])
        eps21 = ComplexMatrixSlice(coeff4, [0,1], [2])
        eps12 = ComplexMatrixSlice(coeff4, [2], [0,1])
        eps22 = ComplexMatrixSlice(coeff4, [2], [2])
        '''
        eps11 = coeff4[[0,1], [0,1]]
        eps21 = coeff4[[0,1], 2]
        eps12 = coeff4[2, [0,1]]
        eps22 = coeff4[2,2]
        
        eps = [eps11, eps12, eps21, eps22]

        '''
        tmp = ComplexMatrixInv(coeff2)
        '''
        tmp = coeff2.inv()

        '''
        mu11 = ComplexMatrixSlice(tmp, [0,1], [0,1])   #-> mul1 = tmp[[0,1], [0,1]]
        mu11 = ComplexMatrixAdj(mu11)                  #-> mul1 = mul1.adj()
        mu22 = ComplexMatrixSlice(tmp, [2], [2])                        
        k2_over_mu11 =   ComplexMatrixProduct(mu11, kz*kz)  #-> k2_over_mul1 = mul1*(kz*kz)
        ik_over_mu11 =   ComplexMatrixProduct(mu11, 1j*kz)
        '''
        mu11 = tmp[[0,1], [0,1]]
        mu11 = mu11.adj()
        mu22 = tmp[2,2]
        k2_over_mu11 = mu11*(kz*kz)
        ik_over_mu11 = mu11*(1j*kz)
        
        mu = [mu11, mu22, k2_over_mu11, ik_over_mu11]

        return eps, mu, kz
            
    def add_bf_contribution(self, engine, a, real = True, kfes=0):
        #freq, omega = self.get_root_phys().get_freq_omega()
        eps, mu, kz =  self.get_coeffs_2()
        
        self.set_integrator_realimag_mode(real)

        if real:
            if kfes == 0:           
                dprint1("Add ND contribution(real)" + str(self._sel_index))
            elif kfes == 1:
                dprint1("Add H1 contribution(real)" + str(self._sel_index))               
        else:
            if kfes == 0:           
                dprint1("Add ND contribution(imag)" + str(self._sel_index))
            elif kfes == 1:
                dprint1("Add H1 contribution(imag)" + str(self._sel_index))               
        
        self.call_bf_add_integrator(eps,  mu, kz,  engine, a, kfes)
        
        '''
        if kfes == 0: ## ND element (Epoloidal)
            if real:       
                dprint1("Add ND contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add ND contribution(imag)" + str(self._sel_index))
         
            
            self.add_integrator(engine, 'mur', mu[1],
                                a.AddDomainIntegrator,
                                mfem.CurlCurlIntegrator)
            self.add_integrator(engine, 'epsilon_sigma', eps[0],
                                a.AddDomainIntegrator,
                                mfem.VectorFEMassIntegrator)
            
            if kz != 0:
                self.add_integrator(engine, 'mur', mu[2],
                                    a.AddDomainIntegrator,
                                    mfem.VectorFEMassIntegrator)
                
        elif kfes == 1: ## H1 element (Etoroidal)
            self.add_integrator(engine, 'mur', mu[0],
                                a.AddDomainIntegrator,
                                mfem.DiffusionIntegrator)
            self.add_integrator(engine, 'epsilon_sigma', eps[3],
                                a.AddDomainIntegrator,
                                mfem.MassIntegrator)
        else:
            pass
        '''
    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real = True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) +')'
                    +str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) +')'
                    +str(self._sel_index))
       
        #freq, omega = self.get_root_phys().get_freq_omega()
        eps, mu, kz =  self.get_coeffs_2()

        self.set_integrator_realimag_mode(real)        
        self.call_mix_add_integrator(eps, mu, engine, mbf, r, c, is_trans)

        '''
        if r == 1 and c == 0:        
            #if  is_trans:
            # (a_vec dot u_vec, v_scalar)                        
            itg = mfem.MixedDotProductIntegrator
            self.add_integrator(engine, 'epsilon', eps[1],
                                mbf.AddDomainIntegrator, itg)

            # (-a u_vec, div v_scalar)            
            itg =  mfem.MixedVectorWeakDivergenceIntegrator
            self.add_integrator(engine, 'mur', mu[3],
                                mbf.AddDomainIntegrator, itg)
            #print r, c, mbf
        else:
            # (a_vec dot u_scalar, v_vec)
            itg = mfem.MixedVectorProductIntegrator
            self.add_integrator(engine, 'epsilon', eps[2],
                             mbf.AddDomainIntegrator, itg)
            
            # (a grad u_scalar, v_vec)
            itg =  mfem.MixedVectorGradientIntegrator
            self.add_integrator(engine, 'mur', mu[3],
                             mbf.AddDomainIntegrator, itg)
        '''
    def add_domain_variables(self, v, n, suffix, ind_vars):

        from petram.helper.variables import add_constant
        
        e, m, s, kz = self.vt.make_value_or_expression(self)

        self.do_add_matrix_expr(v, suffix, ind_vars, 'epsilonr', e)
        self.do_add_matrix_expr(v, suffix, ind_vars, 'mur', m)
        self.do_add_matrix_expr(v, suffix, ind_vars, 'sigma', s)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')
        
        add_constant(v, 'kz', suffix, np.float64(kz),
                     domains = self._sel_index,
                     gdomain = self._global_ns)


    



'''
   Surface Current Boundary Condition

   On external surface    n \times H = J_surf
   On internal surface    n \times (H1 - H2) = J_surf (not tested)

   (note)
    In MKSA,  1/mu curl E = -dB/dt 1/mu= i omega B / mu .
      n \times 1/mu curl E = n \times -dB/dt 1/mu
                           = n \times i omega H 
                           = i omega J_surf
    
 
    Therefore, surface integral 
        \int W \dot (n \times 1/mu curl E) d\Omega
    becomes
         \int W \dot (i omega J_surf) d\Omega

    and VectorFEDomainLFIntegrator can be used.


   CopyRight (c) 2016-  S. Shiraiwa
'''
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.coefficient import VCoeff
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.em3d.em3d_base import EM3D_Bdry

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_SurfJ')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('surfJ', VtableElement('surfJ', type='complex',
                                guilabel='Surface J',
                                suffix=('x', 'y', 'z'),
                                default=[0, 0, 0],
                                tip="surface current")),)
'''
from petram.phys.phys_model  import Phys, VectorPhysCoefficient
class Jsurf(VectorPhysCoefficient):
   def __init__(self, *args, **kwargs):
       omega = kwargs.pop('omega', 1.0)
       self.mur = kwargs.pop('mur', 1.0)
       self.fac = -1j*omega
       super(Jsurf, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(Jsurf, self).EvalValue(x)
       v = self.fac * v
       if self.real:  return v.real
       else: return v.imag
'''


def JsurfCoeff(exprs, ind_vars, l, g, omega):
    fac = -1j * omega
    coeff = VCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def bdry_constraints():
    return [EM3D_SurfJ]


class EM3D_SurfJ(EM3D_Bdry):
    is_essential = False
    vt = Vtable(data)

    def has_lf_contribution(self, kfes=0):
        if kfes != 0:
            return False
        return True

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if kfes != 0:
            return
        if real:
            dprint1("Add LF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + str(self._sel_index))

        freq, omega = self.get_root_phys().get_freq_omega()
        f_name = self.vt.make_value_or_expression(self)

        '''
        coeff1 = Jsurf(3, f_name[0],  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real = real, omega = omega,)
        '''
        self.set_integrator_realimag_mode(real)
        coeff1 = JsurfCoeff(f_name[0],  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns, omega)

        self.add_integrator(engine, 'surfJ', coeff1,
                            b.AddBoundaryIntegrator,
                            mfem.VectorFEDomainLFIntegrator)



'''
   Cold plasma:
'''
from petram.phys.common.rf_dispersion_coldplasma import (stix_options,
                                                         default_stix_option,
                                                         default_stix_modelvalue,
                                                         col_model_options,
                                                         default_col_model,
                                                         vtable_data0)

from petram.phys.phys_const import mu0, epsilon0
from petram.phys.numba_coefficient import (func_to_numba_coeff_scalar,
                                           func_to_numba_coeff_vector,
                                           func_to_numba_coeff_matrix)
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.phys_model import PhysCoefficient, VectorPhysCoefficient
from petram.phys.phys_model import MatrixPhysCoefficient, Coefficient_Evaluator
from petram.phys.em2da.em2da_base import EM2Da_Bdry, EM2Da_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2Da_ColdPlasma')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

vtable_data = vtable_data0.copy()
vtable_data.extend([
    ('t_mode', VtableElement('t_mode', type="float",
                             guilabel='m',
                             default=0.0,
                             tip="mode number")), ])


'''
Expansion of matrix is as follows

               [e_rz  e_12 ][Erz ]
[Wrz, Wphx] =  [           ][    ] = Wrz e_rz Erz + Wrz e_12 Ephi 
               [e_21  e_phi][Ephx]

                                + Wphi e_21 Erz + Wphi*e_phi*Ephi


  Erz = Er e_r + Ez e_z
  Ephx = rho Ephi
'''


def domain_constraints():
    return [EM2Da_ColdPlasma]


class EM2Da_ColdPlasma(EM2Da_Domain):
    vt = Vtable(vtable_data)

    def has_bf_contribution(self, kfes):
        if kfes == 0:
            return True
        elif kfes == 1:
            return True
        else:
            return False

    def has_mixed_contribution(self):
        return True

    def get_mixedbf_loc(self):
        '''
        r, c, and flag1, flag2 of MixedBilinearForm
           flag1 : take transpose
           flag2 : take conj
        '''
        return [(0, 1, 1, 1), (1, 0, 1, 1), ]  # (0, 1, -1, 1)]

    def attribute_set(self, v):
        EM2Da_Domain.attribute_set(self, v)
        v["stix_terms"] = default_stix_modelvalue
        v["col_model"] = default_col_model
        return v

    def config_terms(self, evt):
        from petram.phys.common.rf_stix_terms_panel import ask_rf_stix_terms

        self.vt.preprocess_params(self)
        _B, _dens_e, _t_e, _dens_i, _masses, charges, _tmode = self.vt.make_value_or_expression(
            self)

        num_ions = len(charges)
        win = evt.GetEventObject()
        value = ask_rf_stix_terms(win, num_ions, self.stix_terms)
        self.stix_terms = value

    def stix_terms_str(self):
        from petram.phys.common.rf_dispersion_coldplasma import value2modelstr
        return value2modelstr(self.stix_terms)

    def panel1_param(self):
        panels = super(EM2Da_ColdPlasma, self).panel1_param()
        panels.append(["col. model", None, 1, {"values": col_model_options}])
        panels.extend([["Stix terms", "", 2, None],
                       [None, None, 341, {"label": "Customize terms",
                                          "func": "config_terms",
                                          "sendevent": True,
                                          "noexpand": True}], ])

        return panels

    def get_panel1_value(self):
        values = super(EM2Da_ColdPlasma, self).get_panel1_value()
        values.extend([self.col_model, self.stix_terms_str(), self])
        return values

    def import_panel1_value(self, v):
        check = super(EM2Da_ColdPlasma, self).import_panel1_value(v[:-3])
        self.col_model = v[-3]
        return check

    @property
    def jited_coeff(self):
        return self._jited_coeff

    def compile_coeffs(self):
        self._jited_coeff = self.get_coeffs()

    def get_coeffs(self):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        B, dens_e, t_e, dens_i, masses, charges, tmode = self.vt.make_value_or_expression(
            self)
        ind_vars = self.get_root_phys().ind_vars
        col_model = self.col_model

        from petram.phys.common.rf_dispersion_coldplasma import build_coefficients
        coeff1, coeff2, coeff3, coeff_nuei = build_coefficients(ind_vars, omega, B, dens_e, t_e,
                                                                        dens_i, masses, charges, col_model, cnorm,
                                                                        self._global_ns, self._local_ns,
                                                                        sdim=2, terms=self.stix_terms)

        return coeff1, coeff2, coeff3, coeff_nuei, tmode

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        coeff1, coeff2, coeff3, _coeff_nuei, tmode = self.jited_coeff
        self.set_integrator_realimag_mode(real)

        invmu = coeff2.inv()
        coeff4 = coeff1 + coeff3

        eps11 = coeff4[[0, 2], [0, 2]]
        eps22 = coeff4[1, 1]

        def invmu_x_r(ptx, invmu):
            return invmu[0, 0]*ptx[0]

        def invmu_o_r(ptx, invmu):
            return invmu[0, 0]/ptx[0]

        def invmu_o_r_tt(ptx, invmu):
            return invmu[0, 0]/ptx[0]*tmode*tmode

        def eps11_x_r(ptx, eps11_):
            return eps11_*ptx[0]

        def eps22_o_r(ptx, eps22_):
            return eps22_/ptx[0]

        if kfes == 0:  # ND element (Epoloidal)
            if real:
                dprint1("Add ND contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add ND contribution(imag)" + str(self._sel_index))

            se_x_r = func_to_numba_coeff_matrix(eps11_x_r, shape=(2, 2),
                                                complex=True,
                                                dependency=(eps11,))
            imu_x_r = func_to_numba_coeff_scalar(invmu_x_r,
                                                 complex=True,
                                                 dependency=(invmu,))

            self.add_integrator(engine, 'mur', imu_x_r,
                                a.AddDomainIntegrator,
                                mfem.CurlCurlIntegrator)
            self.add_integrator(engine, 'sigma_epsilonr', se_x_r,
                                a.AddDomainIntegrator,
                                mfem.VectorFEMassIntegrator)

            if tmode != 0:
                imu_o_r_2 = func_to_numba_coeff_scalar(invmu_o_r_tt,
                                                       complex=True,
                                                       dependency=(invmu,))

                self.add_integrator(engine, 'mur', imu_o_r_2,
                                    a.AddDomainIntegrator,
                                    mfem.VectorFEMassIntegrator)

        elif kfes == 1:  # H1 element (Etoroidal)
            if real:
                dprint1("Add H1 contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add H1 contribution(imag)" + str(self._sel_index))

            se_o_r = func_to_numba_coeff_scalar(eps22_o_r,
                                                complex=True,
                                                dependency=(eps22,))
            imv_o_r_1 = func_to_numba_coeff_scalar(invmu_o_r,
                                                   complex=True,
                                                   dependency=(invmu,))

            self.add_integrator(engine, 'mur', imv_o_r_1,
                                a.AddDomainIntegrator,
                                mfem.DiffusionIntegrator)
            self.add_integrator(engine, 'epsilonr', se_o_r,
                                a.AddDomainIntegrator,
                                mfem.MassIntegrator)
        else:
            pass

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real=True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))

        coeff1, coeff2, coeff3, _coeff_nuei, tmode = self.jited_coeff
        self.set_integrator_realimag_mode(real)

        invmu = coeff2.inv()
        coeff4 = coeff1 + coeff3

        eps12 = coeff4[[0, 2], 1]
        eps21 = coeff4[1, [0, 2]]

        def iinvmu_o_r_t(ptx, invmu):
            return 1j*invmu[0, 0]/ptx[0]*tmode
        imv_o_r_3 = func_to_numba_coeff_scalar(iinvmu_o_r_t,
                                               complex=True,
                                               dependency=(invmu,))

        se_21 = eps21
        se_12 = eps12

        if r == 1 and c == 0:

            # if  is_trans:
            # (a_vec dot u_vec, v_scalar)
            itg = mfem.MixedDotProductIntegrator
            self.add_integrator(engine, 'sigma_epsilon_21', se_21,
                                mbf.AddDomainIntegrator, itg)
            # (-a u_vec, div v_scalar)
            itg = mfem.MixedVectorWeakDivergenceIntegrator
            self.add_integrator(engine, 'mur', imv_o_r_3,
                                mbf.AddDomainIntegrator, itg)

        else:
            itg = mfem.MixedVectorProductIntegrator
            self.add_integrator(engine, 'sigma_epsilon_12', se_12,
                                mbf.AddDomainIntegrator, itg)

            # (a grad u_scalar, v_vec)

            itg = mfem.MixedVectorGradientIntegrator
            self.add_integrator(engine, 'mur', imv_o_r_3,
                                mbf.AddDomainIntegrator, itg)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        if len(self._sel_index) == 0:
            return

        freq, omega = self.get_root_phys().get_freq_omega()
        B, dens_e, t_e, dens_i, masses, charges, tmode = self.vt.make_value_or_expression(
            self)
        col_model = self.col_model

        add_constant(v, 'm_mode', suffix, np.float64(tmode),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        from petram.phys.common.rf_dispersion_coldplasma import build_variables

        ss = self.parent.parent.name()+'_'+self.name()  # phys module name + name

        ret = build_variables(v, ss, ind_vars, omega, B, dens_e, t_e,
                              dens_i, masses, charges, col_model,
                              self._global_ns, self._local_ns,
                              terms=self.stix_terms)

        from petram.phys.common.rf_dispersion_coldplasma import add_domain_variables_common
        add_domain_variables_common(self, ret, v, suffix, ind_vars)

        var = ['r', 'phi', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')

        return




'''
EM2D : 2D Frequency domain Maxwell equation.

    This module is meant to solve 

    (curl v, curl u) + (v, e u) 
           - (v, n x (1/mu curl u)_s = i w (jext, u) 

    We decompose the x-y-z filed coponents in in-plane and out-of-plane 
    components, Exy and Ez (respectively).  

    Exy = (Ex, Ey) and  where Exy is Nedelec and Ez is H1

    We use
        curl E = curl Exy + ez x (ik Exy - grad Ez), 
    where k is out-of-plane wave number

    curl-curl is decomposed to 
       (curl Exy, curl Wxy) + (ik Exy - grad Ez,  ik Wxy - grad Wz),

    mass 
       e ((Exy, Wxy) + (Ez, Wz)), 

    where e = - epsion * w^2 - i * sigma * w, w is frequency,
    and  mu, epislon, and simga is regular EM parameters.

    ( , ) is volume integral and ( , )_s is surface intengral
    on non-essential boundaries.

    exp(-i w t) is assumed.

 *sigma
  Domain:   
     EM2D_Anisotropic : tensor dielectric
     EM2D_Vac         : scalar dielectric
     EM2D_ExtJ        : external current

  Boundary:
     EM2D_PEC         : Perfect electric conductor
     EM2D_PMC         : Perfect magnetic conductor
     EM2D_H           : Mangetic field boundary (N.I)
     EM2D_SurfJ       : Surface current         (N.I)
     EM2D_Port        : TE, TEM, Coax port      (N.I)
     EM2D_E           : Electric field
     EM2D_Continuity  : Continuitiy
     EM2D_Z           : Impedance

'''
import numpy as np
import traceback

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys
from petram.phys.common.em_base import EMPhysModule
from petram.phys.em2d.em2d_base import EM2D_Bdry
from petram.phys.em2d.em2d_vac import EM2D_Vac

from petram.phys.vtable import VtableElement, Vtable

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2DModel')


class EM2D_DefDomain(EM2D_Vac):
    can_delete = False
    nlterms = []
    # vt  = Vtable(data1)
    # do not use vtable here, since we want to use
    # vtable defined in EM2D_Vac in add_bf_conttribution

    def __init__(self, **kwargs):
        super(EM2D_DefDomain, self).__init__(**kwargs)

    def panel1_param(self):
        return [['Default Domain (Vac)',   "eps_r=1, mu_r=1, sigma=0",  2, {}], ]

    def get_panel1_value(self):
        return ["eps_r=1, mu_r=1, sigma=0", ]

    def import_panel1_value(self, v):
        pass

    def panel1_tip(self):
        return None

    def get_possible_domain(self):
        return []

    def get_possible_child(self):
        return self.parent.get_possible_domain()


data2 = (('label1', VtableElement(None,
                                  guilabel='Default Bdry (PMC)',
                                  default="Ht = 0",
                                  tip="this is a natural BC")),)


class EM2D_DefBdry(EM2D_Bdry):
    can_delete = False
    is_essential = False
    nlterms = []
    vt = Vtable(data2)

    def __init__(self, **kwargs):
        super(EM2D_DefBdry, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM2D_DefBdry, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = ['remaining']
        return v

    def get_possible_bdry(self):
        return []


class EM2D_DefPair(Pair, Phys):
    can_delete = False
    is_essential = False
    is_complex = True

    def __init__(self, **kwargs):
        super(EM2D_DefPair, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM2D_DefPair, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_possible_pair(self):
        return []


class EM2D(EMPhysModule):
    der_vars_base = ['Bx', 'By', 'Bz']
    der_vars_vec = ['E', 'B']
    geom_dim = 2

    def __init__(self, **kwargs):
        super(EM2D, self).__init__(**kwargs)
        self['Domain'] = EM2D_DefDomain()
        self['Boundary'] = EM2D_DefBdry()
        self['Pair'] = EM2D_DefPair()

    @property
    def dep_vars(self):
        '''
        list of dependent variables, for example.
           [Et, rEf]      
           [Et, rEf, psi]
        '''
        ret = self.dep_vars_base
        return [x + self.dep_vars_suffix for x in ret]

    @property
    def dep_vars0(self):
        '''
        list of dependent variables, for example.
           [Exy, Ez]      
           [Exy, Ez, psi]
        '''
        ret = self.dep_vars_base
        return [x + self.dep_vars_suffix for x in ret]

    @property
    def dep_vars_base(self):
        if self._has_div_constraint():
            ret = ['Exy', 'Ez', 'psi']
        else:
            ret = ['Exy', 'Ez']
        return ret

    def get_fec_type(self, idx):
        values = ['ND', 'H1', 'H1']
        return values[idx]

    def get_fec(self):
        v = self.dep_vars
        if len(v) == 2:  # normal case
            return [(v[0], 'ND_FECollection'),
                    (v[1], 'H1_FECollection'), ]
        else:            # with divergence constraints
            return [(v[0], 'ND_FECollection'),
                    (v[1], 'H1_FECollection'),
                    (v[2], 'H1_FECollection'), ]

    def _has_div_constraint(self):
        return False

    def attribute_set(self, v):
        v = super(EM2D, self).attribute_set(v)
        v["element"] = 'ND_FECollection, H1_FECollection'
        v["ndim"] = 2
        v["ind_vars"] = 'x, y'
        v["dep_vars_suffix"] = ''
        return v

    def panel1_param(self):
        panels = super(EM2D, self).panel1_param()
        panels.extend([  # self.make_param_panel('freq',  self.freq_txt),
            ["independent vars.", self.ind_vars, 0, {}],
            ["dep. vars. suffix", self.dep_vars_suffix, 0, {}],
            ["dep. vars.", ','.join(self.dep_vars), 2, {}],
            ["ns vars.", ','.join(EM2D.der_vars_base), 2, {}], ])

        return panels

    def get_panel1_value(self):
        names = ', '.join([x for x in self.dep_vars])
        names2 = ', '.join(list(self.get_default_ns()))
        val = super(EM2D, self).get_panel1_value()
        # val.extend([self.freq_txt, self.ind_vars, self.dep_vars_suffix,
        #            names, names2, ])
        val.extend([self.ind_vars, self.dep_vars_suffix,
                    names, names2, ])
        return val

    def import_panel1_value(self, v):
        v = super(EM2D, self).import_panel1_value(v)
        # self.freq_txt = str(v[0])
        self.ind_vars = str(v[0])
        self.dep_vars_suffix = str(v[1])

    def get_possible_bdry(self):
        if EM2D._possible_constraints is None:
            self._set_possible_constraints('em2d')
        bdrs = super(EM2D, self).get_possible_bdry()
        return EM2D._possible_constraints['bdry'] + bdrs

    def get_possible_domain(self):
        if EM2D._possible_constraints is None:
            self._set_possible_constraints('em2d')

        doms = super(EM2D, self).get_possible_domain()
        return EM2D._possible_constraints['domain'] + doms

    def get_possible_edge(self):
        return []

    def get_possible_pair(self):
        if EM2D._possible_constraints is None:
            self._set_possible_constraints('em2d')

        pairs = super(EM2D, self).get_possible_pair()
        return EM2D._possible_constraints['pair'] + pairs

    def get_possible_point(self):
        return []

    def add_variables(self, v, name, solr, soli=None):
        from petram.helper.variables import add_coordinates
        from petram.helper.variables import add_scalar
        from petram.helper.variables import add_components
        from petram.helper.variables import add_elements
        from petram.helper.variables import add_expression
        from petram.helper.variables import add_component_expression as addc_expression
        from petram.helper.variables import add_surf_normals
        from petram.helper.variables import add_constant

        from petram.phys.em2da.eval_deriv import eval_curl, eval_grad

        v = super(EM2D, self).add_variables(v, name, solr, soli)

        def eval_curlExy(gfr, gfi=None):
            gfr, gfi, extra = eval_curl(gfr, gfi)
            return gfr, gfi, extra

        def eval_gradEz(gfr, gfi=None):
            gfr, gfi, extra = eval_grad(gfr, gfi)
            return gfr, gfi, extra

        ind_vars = [x.strip() for x in self.ind_vars.split(',')]
        suffix = self.dep_vars_suffix

        add_coordinates(v, ind_vars)
        add_surf_normals(v, ind_vars)

        if name.startswith('Exy'):
            add_elements(v, 'E', suffix, ind_vars, solr, soli, elements=[0, 1])
            add_scalar(v, 'curlExy', suffix, ind_vars, solr, soli,
                       deriv=eval_curlExy)
            addc_expression(v, 'B', suffix, ind_vars,
                            '-1j/omega*curlExy', ['curlExy', 'omega'], 'z')

        elif name.startswith('Ez'):
            add_scalar(v, 'Ez', suffix, ind_vars, solr, soli, vars=['E'])
            add_components(v, 'gradE', suffix, ind_vars, solr, soli,
                           deriv=eval_gradEz)

        elif name.startswith('psi'):
            add_scalar(v, 'psi', suffix, ind_vars, solr, soli)

        add_expression(v, 'E', suffix, ind_vars, 'array([Ex, Ey, Ez])',
                       ['Ex', 'Ey', 'Ez'])

        addc_expression(v, 'B', suffix, ind_vars,
                        '-1j/omega*(-1j*kz*Ey + gradEy)',
                        ['m_mode', 'E', 'omega'], 0)
        addc_expression(v, 'B', suffix, ind_vars,
                        '-1j/omega*(1j*kz*Ex - gradEx)',
                        ['m_mode', 'E', 'omega'], 1)
        add_expression(v, 'B', suffix, ind_vars,
                       'array([Bx, By, Bz])',
                       ['B'])

        # Poynting Flux

        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ey)*Bz - conj(Ez)*By)/mu0',
                        ['B', 'E'], 0)
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ez)*Bx - conj(Ex)*Bz)/mu0',
                        ['B', 'E'], 'y')
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ex)*By - conj(Ey)*Bx)/mu0',
                        ['B', 'E'], 'z')

        return v

    def get_fes_for_dep(self, unknown_name, soldict):
        keys = soldict.keys()
        for k in keys:
            if unknown_name.startswith('phi'):
                if k.startswith('psi'):
                    break
            elif unknown_name.startswith('Et'):
                if k.startswith('Exy'):
                    break
            else:
                if k.startswith('Ez'):
                    break
        sol = soldict[k]
        solr = sol[0]
        soli = sol[1] if len(sol) > 1 else None
        return solr, soli



'''
   external current source
'''
from petram.phys.coefficient import VCoeff
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.em3d.em3d_base import EM3D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_extJ')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('jext', VtableElement('jext', type='complex',
                               guilabel='External J',
                               suffix=('x', 'y', 'z'),
                               default=[0, 0, 0],
                               tip="volumetric external current")),)

'''
from petram.phys.phys_model  import Phys, VectorPhysCoefficient
class Jext(VectorPhysCoefficient):
   def __init__(self, *args, **kwargs):
       self.omega = kwargs.pop('omega', 1.0)
       super(Jext, self).__init__(*args, **kwargs)
   def EvalValue(self, x):
       v = super(Jext, self).EvalValue(x)
       v = 1j * self.omega * v
       if self.real:  return v.real
       else: return v.imag
'''


def JextCoeff(exprs, ind_vars, l, g, omega):
    fac = 1j * omega
    coeff = VCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def domain_constraints():
    return [EM3D_ExtJ]


class EM3D_ExtJ(EM3D_Domain):
    is_secondary_condition = True
    has_3rd_panel = False
    vt = Vtable(data)

    def has_lf_contribution(self, kfes=0):
        if kfes != 0:
            return False
        return True

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if kfes != 0:
            return
        if real:
            dprint1("Add LF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + str(self._sel_index))

        freq, omega = self.get_root_phys().get_freq_omega()
        f_name = self.vt.make_value_or_expression(self)

        # coeff1 = Jext(3, f_name[0],  self.get_root_phys().ind_vars,
        #                    self._local_ns, self._global_ns,
        #                    real = real, omega = omega)
        self.set_integrator_realimag_mode(real)
        coeff1 = JextCoeff(f_name[0],  self.get_root_phys().ind_vars,
                           self._local_ns, self._global_ns, omega)

        self.add_integrator(engine, 'jext', coeff1,
                            b.AddDomainIntegrator,
                            mfem.VectorFEDomainLFIntegrator)

        '''
        coeff1 = self.restrict_coeff(coeff1, engine, vec = True)
        b.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(coeff1))
        '''



#
#  subroutine for analyzing port goemetry
#
from petram.helper.geom import find_circle_center_radius
from petram.helper.geom import connect_pairs
from petram.model import Bdry
from petram.mfem_config import use_parallel

import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('RFPortGeomgetry')

if use_parallel:
    import mfem.par as mfem
    from mfem.common.mpi_debug import nicePrint
    from petram.helper.mpi_recipes import *
else:
    import mfem.ser as mfem
    nicePrint = dprint1


def analyze_geom_norm_ref(mesh, ibe, ref):

    el = mesh.GetBdrElement(ibe[0])
    Tr = mesh.GetBdrElementTransformation(ibe[0])
    rules = mfem.IntegrationRules()
    ir = rules.Get(el.GetGeometryType(), 1)
    Tr.SetIntPoint(ir.IntPoint(0))
    nor = mfem.Vector(3)
    mfem.CalcOrtho(Tr.Jacobian(), nor)

    norm = nor.GetDataArray().copy()
    norm = norm / np.sqrt(np.sum(norm**2))

    if ref.strip() == "":
        ref_ptx = None
    else:
        iptx = mesh.extended_connectivity['vert2vert'][int(ref)]
        ref_ptx = mesh.GetVertexArray(iptx)

    return norm, ref_ptx


def analyze_rect_geom(port_model, mesh, ibe, norm):
    data = {}

    edges = np.array([mesh.GetBdrElementEdges(i)[0]
                      for i in ibe]).flatten()
    d = {}
    for x in edges:
        d[x] = x in d
    edges = [x for x in d.keys() if not d[x]]
    ivert = [mesh.GetEdgeVertices(x) for x in edges]
    ivert = connect_pairs(ivert)
    vv = np.vstack([mesh.GetVertexArray(i) for i in ivert])

    data["ctr"] = (np.max(vv, 0) + np.min(vv, 0)) / 2.0
    dprint1("Center " + list(data["ctr"]).__repr__())

    # rectangular port
    # idx = np.argsort(np.sqrt(np.sum((vv - data["ctr"])**2,1)))
    # corners = vv[idx[-4:],:]
    # since vv is cyclic I need to omit last one element here..
    idx = np.argsort(np.sqrt(np.sum((vv[:-1] - data["ctr"])**2, 1)))
    corners = vv[:-1][idx[-4:], :]
    for i in range(4):
        dprint1("Corner " + list(corners[i]).__repr__())
    tmp = np.sort(np.sqrt(np.sum((corners - corners[0, :])**2, 1)))
    data["b"] = tmp[1]
    data["a"] = tmp[2]
    tmp = np.argsort(np.sqrt(np.sum((corners - corners[0, :])**2, 1)))
    data["c"] = corners[0]  # corner
    data["b_vec"] = corners[tmp[1]] - corners[0]
    data["a_vec"] = np.cross(data["b_vec"], norm)
    #            data["a"]_vec = corners[tmp[2]]-corners[0]
    data["b_vec"] = data["b_vec"] / np.sqrt(np.sum(data["b_vec"]**2))
    data["a_vec"] = data["a_vec"] / np.sqrt(np.sum(data["a_vec"]**2))
    if np.sum(np.cross(data["a_vec"], data["b_vec"]) * norm) > 0:
        data["a_vec"] = -data["a_vec"]

    if port_model.mode == 'TEM':
        '''
        special handling
        set a vector along PEC-like edge, regardless the actual
        length of edges
        '''
        for i in range(nbe):
            if (edges[0] in mesh.GetBdrElementEdges(i)[0] and
                    port_model._sel_index[0] != mesh.GetBdrAttribute(i)):
                dprint1("Checking surface :", mesh.GetBdrAttribute(i))
                attr = mesh.GetBdrAttribute(i)
                break
        for node in port_model.get_root_phys().walk():
            if not isinstance(node, Bdry):
                continue
            if not node.enabled:
                continue
            if attr in node._sel_index:
                break
        from petram.model import Pair
        ivert = mesh.GetEdgeVertices(edges[0])
        vect = mesh.GetVertexArray(
            ivert[0]) - mesh.GetVertexArray(ivert[1])
        vect = vect / np.sqrt(np.sum(vect**2))
        do_swap = False
        if (isinstance(node, Pair) and
                np.abs(np.sum(data["a_vec"] * vect)) > 0.9):
            do_swap = True
        if (not isinstance(node, Pair) and
                np.abs(np.sum(data["a_vec"] * vect)) < 0.001):
            do_swap = True
        if do_swap:
            dprint1("swapping port edges")
            tmp = data["a_vec"]
            data["a_vec"] = -data["b_vec"]
            data["b_vec"] = tmp
            # - sign is to keep a \times b direction.
            tmp = data["a"]
            data["a"] = data["b"]
            data["b"] = tmp
    if data["a_vec"][np.argmax(np.abs(data["a_vec"]))] < 0:
        data["a_vec"] = -data["a_vec"]
        data["b_vec"] = -data["b_vec"]
    dprint1("Long Edge  " + data["a"].__repr__())
    dprint1("Long Edge Vec." + list(data["a_vec"]).__repr__())
    dprint1("Short Edge  " + data["b"].__repr__())
    dprint1("Short Edge Vec." + list(data["b_vec"]).__repr__())

    return data, vv


def analyze_coax_geom(mesh, ibe, norm, ref_ptx):
    data = {}

    edges = np.array([mesh.GetBdrElementEdges(i)[0]
                      for i in ibe]).flatten()
    d = {}
    for x in edges:
        d[x] = x in d
    edges = [x for x in d.keys() if not d[x]]
    ivert = [mesh.GetEdgeVertices(x) for x in edges]
    iv1, iv2 = connect_pairs(ivert)  # index of outer/inner circles
    vv1 = np.vstack([mesh.GetVertexArray(i) for i in iv1])
    vv2 = np.vstack([mesh.GetVertexArray(i) for i in iv2])
    ctr1, a1 = find_circle_center_radius(vv1, norm)
    ctr2, b1 = find_circle_center_radius(vv2, norm)
    data["ctr"] = np.mean((ctr1, ctr2), 0)
    data["a"] = a1 if a1 < b1 else b1
    data["b"] = a1 if a1 > b1 else b1

    ax1 = ref_ptx - data["ctr"]
    ax1 = ax1/np.sqrt(np.sum(ax1**2))
    ax2 = np.cross(norm, data["ctr"])
    data["ax1"] = ax1
    data["ax2"] = ax2

    dprint1("Big R:  " + data["b"].__repr__())
    dprint1("Small R: " + data["a"].__repr__())
    dprint1("Center:  " + data["ctr"].__repr__())
    dprint1("Axis1:  " + data["ax1"].__repr__())
    dprint1("Axis2:  " + data["ax2"].__repr__())
    return data, vv1


def analyze_circular_geom(mesh, ibe, norm, ref_ptx):
    data = {}

    edges = np.array([mesh.GetBdrElementEdges(i)[0]
                      for i in ibe]).flatten()
    d = {}
    for x in edges:
        d[x] = x in d
    edges = [x for x in d.keys() if not d[x]]
    ivert = [mesh.GetEdgeVertices(x) for x in edges]
    iv1 = connect_pairs(ivert)  # index of outer/inner circles
    vv1 = np.vstack([mesh.GetVertexArray(i) for i in iv1])
    ctr, a1 = find_circle_center_radius(vv1, norm)
    data["ctr"] = ctr
    data["a"] = a1

    ax1 = ref_ptx - data["ctr"]
    ax1 = ax1/np.sqrt(np.sum(ax1**2))
    ax2 = np.cross(norm, ax1)
    data["ax1"] = ax1
    data["ax2"] = ax2

    dprint1("Radius:  " + data["a"].__repr__())
    dprint1("Center:  " + data["ctr"].__repr__())
    dprint1("Axis1:  " + data["ax1"].__repr__())
    dprint1("Axis2:  " + data["ax2"].__repr__())

    return data, vv1




'''
EC3D : Electric Circilt

   This module is meant to solve 

 *sigma
  Domain:   
     EC3D_Conductor   : scalar dielectric
     EC3D_ExtJ        : external current

  Boundary:
     EC3D_Insulation  : Insulation
     EC3D_Potentail   : Potentai
     EC3D_Ground      : Potentai = 0
     EC3D_Continuity  : Continuitiy

'''
import numpy as np
import traceback

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys, PhysModule
from petram.phys.em3d.em3d_base import EM3D_Bdry
from petram.phys.em3d.em3d_vac import EM3D_Vac

txt_predefined = 'freq, e0, mu0'

from petram.phys.vtable import VtableElement, Vtable

data2 =  (('label1', VtableElement(None, 
                                     guilabel = 'Default Bdry (PMC)',
                                     default =   "Ht = 0",
                                     tip = "this is a natural BC" )),)
class EM3D_DefDomain(EM3D_Vac):
    can_delete = False
    nlterms = []
    #vt  = Vtable(data1)
    #do not use vtable here, since we want to use
    #vtable defined in EM3D_Vac in add_bf_conttribution
    
    def __init__(self, **kwargs):
        super(EM3D_DefDomain, self).__init__(**kwargs)
    def panel1_param(self):
        return [['Default Domain (Vac)',   "eps_r=1, mu_r=1, sigma=0",  2, {}],]
    def get_panel1_value(self):
        return None
    def import_panel1_value(self, v):
        pass
    def panel1_tip(self):
        return None
    def get_possible_domain(self):
        return []
          
        
class EM3D_DefBdry(EM3D_Bdry):
    can_delete = False
    is_essential = False
    nlterms = []          
    vt  = Vtable(data2)                    
    def __init__(self, **kwargs):
        super(EM3D_DefBdry, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM3D_DefBdry, self).attribute_set(v)        
        v['sel_readonly'] = False
        v['sel_index'] = ['remaining']
        return v
    '''   
    def panel1_param(self):
        return [['Default Bdry (PMC)',   "Ht = 0",  2, {}],]

    def get_panel1_value(self):
        return None

    def import_panel1_value(self, v):
        pass
    
    def panel1_tip(self):
        return None
    '''
    def get_possible_bdry(self):
        return []                

class EM3D_DefPair(Pair, Phys):
    can_delete = False
    is_essential = False
    is_complex = True
    def __init__(self, **kwargs):
        super(EM3D_DefPair, self).__init__(**kwargs)
        Phys.__init__(self)
        
    def attribute_set(self, v):
        super(EM3D_DefPair, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_possible_pair(self):
        return []

class EC3D(PhysModule):
    der_var_base = ['Jx' 'Jy', 'Jz']
    geom_dim = 3
    def __init__(self, **kwargs):
        super(EC3D, self).__init__()
        Phys.__init__(self)
        self['Domain'] = EC3D_DefDomain()
        self['Boundary'] = EC3D_DefBdry()
        self['Pair'] = EC3D_DefPair()
        
    @property
    def dep_vars(self):
        '''
        list of dependent variables, for example.
           [V]      
        '''
        ret = ['V']
        return [x + self.dep_vars_suffix for x in ret]            
    @property 
    def dep_vars_base(self):
        return  = ['V']

    def get_fec(self):
        v = self.dep_vars
        return [(v[0], 'H1_FECollection')]
    
    def attribute_set(self, v):
        v = super(EC3D, self).attribute_set(v)
        v["element"] = 'H1_FECollection'
        v["freq_txt"]    = 1.0e9
        v["ind_vars"] = 'x, y, z'
        v["dep_vars_suffix"] = ''
        return v
    
    def panel1_param(self):
        panels = super(EC3D, self).panel1_param()
        panels.extend([
                ["independent vars.", self.ind_vars, 0, {}],
                ["dep. vars. suffix", self.dep_vars_suffix, 0, {}],
                ["dep. vars.", ','.join(self.dep_vars), 2, {}],
                ["derived vars.", ','.join(EC3D.der_var_base), 2, {}],
                ["predefined ns vars.", txt_predefined , 2, {}]])
        return panels
      
    def get_panel1_value(self):
        names  = ','.join([x+self.dep_vars_suffix for x in self.dep_vars])
        names2  = ','.join([x+self.dep_vars_suffix for x in EM3D.der_var_base])
        val =  super(EC3D, self).get_panel1_value()
        val.extend([self.ind_vars, self.dep_vars_suffix,
                    names, names2, txt_predefined])
        return val
    
    def get_panel2_value(self):
        return 'all'
    
    def get_default_ns(self):
        from em3d_const import mu0, epsilon0, q0
        ns =  {'mu0': mu0,
               'e0': epsilon0,
               'q0': q0}
        return ns
    
    def import_panel1_value(self, v):
        v = super(EC3D, self).import_panel1_value(v)
        self.freq_txt =  str(v[0])
        self.ind_vars =  str(v[0])
        self.dep_vars_suffix =  str(v[2])                
        from em3d_const import mu0, epsilon0
        self._global_ns['mu0'] = mu0
        self._global_ns['epsilon0'] = epsilon0
            
    def import_panel2_value(self, v):
        self.sel_index = 'all'
      
    def get_possible_bdry(self):
        from em3d_pec       import EM3D_PEC
        from em3d_pmc       import EM3D_PMC
        from em3d_h       import EM3D_H
        from em3d_surfj       import EM3D_SurfJ
        from em3d_port       import EM3D_Port
        from em3d_e       import EM3D_E
        from em3d_cont       import EM3D_Continuity
        return [EM3D_PEC, EM3D_Port, EM3D_E, EM3D_SurfJ, 
                EM3D_H, EM3D_PMC, EM3D_Continuity]
    
    def get_possible_domain(self):
        from ec3d_domains   import EC3D_Conductor, EC3D_ExtJ

        return [EC3D_Conductor, EM3D_ExtJ]

    def get_possible_edge(self):
        return []                

    def get_possible_pair(self):
        return []

    def get_possible_point(self):
        return []

    def is_complex(self):
        return True

    def add_variables(self, v, name, solr, soli = None):
        from petram.helper.variables import add_coordinates
        from petram.helper.variables import add_scalar
        from petram.helper.variables import add_components
        from petram.helper.variables import add_expression
        from petram.helper.variables import add_surf_normals
        from petram.helper.variables import add_constant      
        
        ind_vars = [x.strip() for x in self.ind_vars.split(',')]
        suffix = self.dep_vars_suffix

        from petram.helper.variables import TestVariable
        #v['debug_test'] =  TestVariable()
        
        add_coordinates(v, ind_vars)        
        add_surf_normals(v, ind_vars)
        
        if name.startswith('E'):
            #add_constant(v, 'freq', suffix, self._global_ns['freq'])
            #add_constant(v, 'mu0',  suffix, self._global_ns['mu0'])
            #add_constant(v, 'e0',  suffix, self._global_ns['e0'])
                           
            add_components(v, 'E', suffix, ind_vars, solr, soli)
            add_components(v, 'B', suffix, ind_vars, solr, soli,
                           deriv=evalB)
            add_expression(v, 'normE', suffix, ind_vars,
                           '(conj(Ex)*Ex + conj(Ey)*Ey +conj(Ez)*Ez)**(0.5)',
                           ['E'])
            add_expression(v, 'normB', suffix, ind_vars,
                           '(conj(Bx)*Bx + conj(By)*By + conj(Bz)*Bz)**(0.5)',
                           ['B'])

            # Poynting Flux
            add_expression(v, 'Poyx', suffix, ind_vars,
                           '(conj(Ey)*Bz - conj(Ez)*By)/mu0', ['B', 'E'])
            add_expression(v, 'Poyy', suffix, ind_vars,
                           '(conj(Ez)*Bx - conj(Ex)*Bz)/mu0', ['B', 'E'])
            add_expression(v, 'Poyz', suffix, ind_vars,
                           '(conj(Ex)*By - conj(Ey)*Bx)/mu0', ['B', 'E'])

            #e = - epsion * w^2 - i * sigma * w
            # Jd : displacement current  = -i omega* e0 er E
            add_expression(v, 'Jdx', suffix, ind_vars,
                           '(-1j*(dot(epsilonr, E))*freq*2*pi*e0)[0]', 
                           ['epsilonr', 'E'])
            add_expression(v, 'Jdy', suffix, ind_vars,
                           '(-1j*(dot(epsilonr, E))*freq*2*pi*e0)[1]', 
                           ['epsilonr', 'E'])
            add_expression(v, 'Jdz', suffix, ind_vars,
                           '(-1j*(dot(epsilonr, E))*freq*2*pi*e0)[2]', 
                           ['epsilonr', 'E'])
            # Ji : induced current = sigma *E
            add_expression(v, 'Jix', suffix, ind_vars,
                           '(dot(sigma, E))[0]', 
                           ['sigma', 'E'])
            add_expression(v, 'Jiy', suffix, ind_vars,
                           '(dot(sigma, E))[1]', 
                           ['sigma', 'E'])
            add_expression(v, 'Jiz', suffix, ind_vars,
                           '(dot(sigma, E))[2]', 
                           ['sigma', 'E'])
            # Jp : polarization current (Jp = -i omega* e0 (er - 1) E
            add_expression(v, 'Jpx', suffix, ind_vars,
                           '(-1j*(dot(epsilonr, E) - E)*freq*2*pi*e0)[0]', 
                           ['epsilonr', 'E'])
            add_expression(v, 'Jpy', suffix, ind_vars,
                           '(-1j*(dot(epsilonr, E) - E)*freq*2*pi*e0)[1]', 
                           ['epsilonr', 'E'])
            add_expression(v, 'Jpz', suffix, ind_vars,
                           '(-1j*(dot(epsilonr, E) - E)*freq*2*pi*e0)[2]', 
                           ['epsilonr', 'E'])
            
            
        elif name.startswith('psi'):
            add_scalar(v, 'psi', suffix, ind_vars, solr, soli)

        return v

               
    def get_fes_for_dep(self, unknown_name, soldict):
        keys = soldict.keys()
        for k in keys:
            if unknown_name.startswith('psi'):
               if k.startswith('psi'): break
            else:
               if k.startswith('E'): break
        sol = soldict[k]
        solr = sol[0]
        soli = sol[1] if len(sol) > 1 else None
        return solr, soli                



from numpy import pi, sin, cos, exp, sqrt, log, arctan2
# constants    
w = 4.6e9 * 2* pi           # omega

e0 = 8.8541878176e-12       # vacuum permittivity

Zi = 1
Zim = 18

q = 1.60217662e-19
qe = q
qi = Zi*q
qim = Zim*q
me = 9.10938356e-31
u = 1.660539040e−27  # unified mass unit
mi = 1.00794*u
mi = 2.01410178*u    # Dutrium
mim = 39.948*u       # Ar (39.948)

imp_frac = 0.01

#poloidal field
from scipy.interpolate import interp2d
Brfit =  interp2d(rgrid, zgrid, brrz)
Bzfit =  interp2d(rgrid, zgrid, btrz)
Btfit =  interp2d(rgrid, zgrid, bzrz)

def Bfields(x, y, z):
    return Brfit([x, y, z]), Bzfit([x, y, z]), Btfit([x, y, z])

def B(x, y, z):
    Br, Bz, Bt = Bfield(x, y, z)
    return sqrt(Br**2+Bz**2+Bt**2)

def ph_th(x, y, z):
    Br, Bz, Bt = Bfield(x, y, z)
    th = arctan2(Bz, Br)
    ph = arctan2(Br*cos(th)+Bz*sin(th),Bt)
    return ph, th

def ne(x, y, z):
    return 1e19
def ni(x, y, z):
    return ne(x, y, z) *(1. - imp_frac)
def nim(x, y, z):
    return ne(x, y, z) * imp_frac)

def Te(x, y, z):
    return 100*q     # 100eV in MKSA 

def Ti(x, y, z):
    return 100*q     # 100eV in MKSA 

def Tim(x, y, z):
    return 100*q     # 100eV in MKSA 

def vTe(x, y, z):
    return sqrt(2*Te(x, y, z)/me)
def vTi(x, y, z):
    return sqrt(2*Ti(x, y, z)/mi)
def vTim(x, y, z):
    return sqrt(2*Tim(x, y, z)/mim)

def LAMBDA(x, y, z):
    # 1+LAMDA as in log(1+LAMBDA) coulomb logarithm
    Te, ne = Te(x, y, z), ne(x, y, z)
    1+12*pi*(e0*Te)**(3./2)/(q**3 * sqrt(ne))

def nu_ei(x, y, z):
    return (qi**2 * qe**2 * ni(x, y, z) *
            log(LAMBDA(x, y, z))/(4 * pi*e0**2*me**2)/vTe(x, y, z)^3)

def nu_eim(x, y, z):
    return (qim**2 * qe**2 * nim(x, y, z) *
            log(LAMBDA(x, y, z))/(4 * pi*e0**2*me**2)/vTe(x, y, z)^3)

def me_eff(x, y z):
    #effective electrons mass (to account for collisions)
    return (1+nu_ei(x, y, z) / 1j/w + nu_eim(x, y, z)/1j/w )*me

def mi_eff(x, y z):
    #effective ion mass (to account for collisions)
    return (1+nu_ei(x, y, z) / 1j/w + nu_eim(x, y, z)/1j/w )*mi

def mim_eff(x, y z):
    #effective ion mass (to account for collisions)
    return (1+nu_ei(x, y, z) / 1j/w + nu_eim(x, y, z)/1j/w )*mim

def wp(x, y, z):
    # electron plasma freuquency
    return sqrt(ne(x, y, z) * qe**2/(me_eff(x, y, z)*e0))

def wpi(x, y, z):
    #sqrt(ni*qi^2/(mi_eff*e0)) ion plasma frequency        
    return sqrt(ni(x, y, z) * qi**2/(mi_eff(x, y, z)*e0))

def wpim(x, y, z):
    # electron plasma freuquency

    return sqrt(nim(x, y, z) * qim**2/(mim_eff(x, y, z)*e0))

def wce(x, y, z):
    # electron cyclotron frequency
    return -q * B(x, y, z)/me_eff(x, y, z)
def wci(x, y, z):
    # ion cyclotron frequency
    return qi * B(x, y, z)/mi_eff(x, y, z)
def wce(x, y, z)
    return qim * B(x, y, z)/mim_eff(x, y, z)

def wc_wp(x, y, z):
    return (wpe(x, y, z), wpi(x, y, z), wpim(x, y, z),
            wce(x, y, z), wci(x, y, z), wcim(x, y, z),)

def P_rf(x, y, z):
    #STIX P
    wpe, wpi, wpim, wce, wci, wcim = wc_wp(x, y, z)    
    return (1 - wpe**2/w**2
              - wpi**2/w**2
              - wpim**2/w**2)
def S_rf(x, y, z):
    #STIX S    
    wpe, wpi, wpim, wce, wci, wcim = wc_wp(x, y, z)
    return 1-wpe**2/(w**2-wce**2)-wpi**2/(w**2-wci**2)-wpim**2/(w**2-wcim**2)

def D_rf(x, y, z):
    # STIX D
    wpe, wpi, wpim, wce, wci, wcim = wc_wp(x, y, z)    
    return (wce*wpe**2/(w*(w**2-wce**2)) + wci*wpi**2/(w*(w**2-wci**2)) +
            wcim*wpim**2/(w*(w**2-wcim**2)))

def Stix(x, y, z):
    return S_rf(x, y, z), D_rf(x, y, z), P_rf(x, y, z)

def e_xx(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return (- P*sin(ph)**2*sin(th)**2
            + P*sin(ph)**2 + S*sin(ph)**2 * sin(th)**2
            - S*sin(ph)**2 + S)
def e_xy(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return (1j*D*sin(th) + P*cos(ph)*cos(th) - S*cos(ph)*cos(th))*sin(ph)

def e_xz(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return -1j*D*cos(ph) + P*sin(ph)**2*sin(th)*cos(th)-S*sin(ph)**2*sin(th)*cos(th)

def e_yx(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return  -(1j*D*sin(th)-P*cos(ph)*cos(th)+S*cos(ph)*cos(th))*sin(ph)

def e_yy(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return P*cos(ph)**2 + S*sin(ph)**2

def e_yz(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return  (1j*D*cos(th)+P*sin(th)*cos(ph)-S*sin(th)*cos(ph))*sin(ph) 

def e_zx(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return 1j*D*cos(ph) + P*sin(ph)**2*sin(th)*cos(th) - S*sin(ph)**2*sin(th)*cos(th)

def e_zy(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return -(1j*D*cos(th)-P*sin(th)*cos(ph)+S*sin(th)*cos(ph))*sin(ph) 

def e_zz(x, y, z):
    ph, th = ph_th(x, y, z)
    S, D, P = Stix(x, y, z)    
    return P*sin(ph)**2*sin(th)**2 - S*sin(ph)**2*sin(th)**2 + S



'''
S S_rf
P P_rf
D D_rf
Ezeta (emw.Er*Br+emw.Ez*Bz+emw.Ephi*Bt)/B
Eeta (emw.Er*Br+emw.Ez*Bz)/Bp*cos(ph)-emw.Ephi*sin(ph)
Bzeta (emw.Br*Br+emw.Bz*Bz+emw.Bphi*Bt)/B
Beta (emw.Br*Br+emw.Bz*Bz)/Bp*cos(ph)-emw.Bphi*sin(ph)

Eb_phi Eb*exp(i*pmode*theta(r,z))*cos(ph+e_angle) not used
Eb_z Eb*exp(i*pmode*theta(r,z))*sin(ph+e_angle)*sin(th) not used
Eb_r Eb*exp(i*pmode*theta(r,z))*sin(ph+e_angle)*cos(th) not used
Bt_toric -Bt
Bp_toric s_toric*Bp
Theta atan2(Bp_toric,Bt_toric) Capital theta in Toric
Eeta_b (emw.Er*cos(th_efit(r,z))+emw.Ez*sin(th_efit(r,z)))*s_toric*cos(Thetab(r,z))-sin(Thetab(r,z))*(-emw.Ephi)
Beta_b (emw.Br*cos(th_efit(r,z))+emw.Bz*sin(th_efit(r,z)))*s_toric*cos(Thetab(r,z))-sin(Thetab(r,z))*(-emw.Bphi)
Ezeta_b (emw.Er*cos(th_efit(r,z))+emw.Ez*sin(th_efit(r,z)))*s_toric*sin(Thetab(r,z))+cos(Thetab(r,z))*(-emw.Ephi)
Bzeta_b (emw.Br*cos(th_efit(r,z))+emw.Bz*sin(th_efit(r,z)))*s_toric*sin(Thetab(r,z))+cos(Thetab(r,z))*(-emw.Bphi)
Epsi (emw.Er*Br+emw.Ez*Bz)/Bp*sin(ph)+emw.Ephi*cos(ph)
tEr_b (emw.Er*cos(th_efit(r,z))+emw.Ez*sin(th_efit(r,z)))*cos(th_efit(r,z))
tEz_b (emw.Er*cos(th_efit(r,z))+emw.Ez*sin(th_efit(r,z)))*sin(th_efit(r,z))
tBr_b (emw.Br*cos(th_efit(r,z))+emw.Bz*sin(th_efit(r,z)))*cos(th_efit(r,z))
tBz_b (emw.Br*cos(th_efit(r,z))+emw.Bz*sin(th_efit(r,z)))*sin(th_efit(r,z))
'''



import traceback
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_Base')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys, PhysModule, VectorPhysCoefficient, PhysCoefficient

# define variable for this BC.
from petram.phys.vtable import VtableElement, Vtable
data =  (('Einit', VtableElement('Einit', type='float',
                                  guilabel = 'E(init)',
                                  suffix =('x', 'y', 'z'),
                                  default = np.array([0,0,0]), 
                                  tip = "initial_E",
                                  chkbox = True)),)

class Einit_xy(VectorPhysCoefficient):
   def EvalValue(self, x):
       v = super(Einit_xy, self).EvalValue(x)
       v = np.array((v[0], v[1]))
       if self.real:  val = v.real
       else: val =  v.imag
       return val
    
class Einit_z(PhysCoefficient):
   def EvalValue(self, x):
       v = super(Einit_z, self).EvalValue(x)
       v = v[2]
       if self.real:  val = v.real
       else: val =  v.imag
       return val

    
class EM2D_Domain(Domain, Phys):
    has_3rd_panel = True    
    vt3  = Vtable(data)   
    def __init__(self, **kwargs):
        super(EM2D_Domain, self).__init__(**kwargs)
        Phys.__init__(self)
        
    def attribute_set(self, v):
        super(EM2D_Domain, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v
    
    def get_init_coeff(self, engine, real=True, kfes=0):
        if kfes > 2: return
        if not self.use_Einit: return
        
        f_name = self.vt3.make_value_or_expression(self)
        if kfes == 0:
            coeff = Einit_xy(2, f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine, vec = True)            
        else:
            coeff = Einit_z(f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine)
         
    def has_pml(self):
        from .em2d_pml import EM2D_PML
        for obj in self.walk():
            if isinstance(obj, EM2D_PML) and obj.enabled:
                return True
             
    def get_pml(self):
        from .em2d_pml import EM2D_PML
        return [obj for obj in self.walk() if isinstance(obj, EM2D_PML) and obj.enabled]
    
    def make_PML_coeff(self, coeff):
        pmls = self.get_pml()
        if len(pmls) > 2: assert False, "Multiple PML is set"
        coeff1 = pmls[0].make_PML_coeff(coeff)
        return coeff1
    
class EM2D_Domain_helper(object): 
   def call_bf_add_integrator(self, eps,  mu, kz, engine, a, kfes):
       if kfes == 0: ## ND element (Exy)
           self.add_integrator(engine, 'mur', mu[1],
                              a.AddDomainIntegrator,
                              mfem.CurlCurlIntegrator)
           self.add_integrator(engine, 'epsilon_sigma', eps[0],
                              a.AddDomainIntegrator,
                              mfem.VectorFEMassIntegrator)

           if kz != 0:
               self.add_integrator(engine, 'mur', mu[2],
                                  a.AddDomainIntegrator,
                                  mfem.VectorFEMassIntegrator)

       elif kfes == 1: ## H1 element (Ez)
           self.add_integrator(engine, 'mur', mu[0],
                               a.AddDomainIntegrator,
                               mfem.DiffusionIntegrator)
           self.add_integrator(engine, 'epsilon_sigma', eps[3],
                               a.AddDomainIntegrator,
                               mfem.MassIntegrator)
       else:
           pass

   def call_mix_add_integrator(self, eps, mu, engine, mbf, r, c, is_trans):
       if r == 1 and c == 0:        
           #if  is_trans:
           # (a_vec dot u_vec, v_scalar)                        
           itg = mfem.MixedDotProductIntegrator
           self.add_integrator(engine, 'epsilon', eps[1],
                                   mbf.AddDomainIntegrator, itg)

           # (-a u_vec, div v_scalar)            
           itg =  mfem.MixedVectorWeakDivergenceIntegrator
           self.add_integrator(engine, 'mur', mu[3],
                                   mbf.AddDomainIntegrator, itg)
       else:
           # (a_vec dot u_scalar, v_vec)
           itg = mfem.MixedVectorProductIntegrator
           self.add_integrator(engine, 'epsilon', eps[2],
                               mbf.AddDomainIntegrator, itg)

           # (a grad u_scalar, v_vec)
           itg =  mfem.MixedVectorGradientIntegrator
           self.add_integrator(engine, 'mur', mu[3],
                              mbf.AddDomainIntegrator, itg)


class EM2D_Bdry(Bdry, Phys):
    has_3rd_panel = True        
    vt3  = Vtable(data)   
    def __init__(self, **kwargs):
        super(EM2D_Bdry, self).__init__(**kwargs)
        Phys.__init__(self)
        
    def attribute_set(self, v):
        super(EM2D_Bdry, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_init_coeff(self, engine, real=True, kfes=0):
        if kfes > 2: return
        if not self.use_Einit: return
        
        f_name = self.vt3.make_value_or_expression(self)
        if kfes == 0:
            coeff = Einit_xy(2, f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine, vec = True)            
        else:
            coeff = Einit_z(f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine)                                    


     




from petram.phys.em2da.em2da_base import EM2Da_Bdry, EM2Da_Domain

from petram.phys.vtable import VtableElement, Vtable   
data =  (('label1', VtableElement(None, 
                                  guilabel = 'Perfect Magnetic Conductor',
                                  default =   "Ht = 0",
                                  tip = "this is a natural BC" )),)
def bdry_constraints():
   return [EM2Da_PMC]

class EM2Da_PMC(EM2Da_Bdry):
    is_essential = False
    nlterms = []
    vt  = Vtable(data)          
    



from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_model import Phys
from petram.model import Domain, Bdry, Pair
from petram.mfem_config import use_parallel
from petram.phys.em3d.em3d_base import EM3D_Bdry, EM3D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_PEC')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('label1', VtableElement(None,
                                 guilabel='Perfect Electric Conductor',
                                 default="Et = 0",
                                 tip="Essential Homogenous BC")),)

def bdry_constraints():
   return [EM3D_PEC]

class EM3D_PEC(EM3D_Bdry):
    has_essential = True
    nlterms = []
    vt = Vtable(data)

    def get_essential_idx(self, kfes):
        if kfes == 0:
            return self._sel_index
        else:
            return []

    def apply_essential(self, engine, gf, kfes, real=False,
                        **kwargs):

        if kfes != 0:
            return
        if real:
            dprint1("Apply Ess. E=0 (real)" + str(self._sel_index))
        else:
            dprint1("Apply Ess. E=0 (imag)" + str(self._sel_index))

        mesh = engine.get_mesh(mm=self)
        ibdr = mesh.bdr_attributes.ToList()
        bdr_attr = [0]*mesh.bdr_attributes.Max()
        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        if kfes == 0:
            coeff1 = mfem.VectorArrayCoefficient(3)
            coeff1.Set(0, mfem.ConstantCoefficient(0.0))
            coeff1.Set(1, mfem.ConstantCoefficient(0.0))
            coeff1.Set(2, mfem.ConstantCoefficient(0.0))
            gf.ProjectBdrCoefficientTangent(coeff1,
                                            mfem.intArray(bdr_attr))



'''
   Vacuum region:
      However, can have arbitrary scalar epsilon_r, mu_r, sigma


'''
from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain, EM2D_Domain_helper
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.coefficient import SCoeff
from petram.mfem_config import use_parallel
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_Vac')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

'''   
from petram.phys.coefficient import PyComplexPowCoefficient as ComplexPow
from petram.phys.coefficient import PyComplexProductCoefficient as ComplexProduct
from petram.phys.coefficient import PyComplexSumCoefficient as ComplexSum

from petram.phys.coefficient import PyComplexMatrixSumCoefficient as ComplexMatrixSum
from petram.phys.coefficient import PyComplexMatrixInvCoefficient as ComplexMatrixInv
from petram.phys.coefficient import PyComplexMatrixSliceCoefficient as ComplexMatrixSlice
from petram.phys.coefficient import PyComplexMatrixProductCoefficient as ComplexMatrixProduct
from petram.phys.coefficient import PyComplexMatrixAdjCoefficient as ComplexMatrixAdj
'''
data = (('epsilonr', VtableElement('epsilonr', type='complex',
                                   guilabel='epsilonr',
                                   default=1.0,
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur', type='complex',
                              guilabel='mur',
                              default=1.0,
                              tip="relative permeability")),
        ('sigma', VtableElement('sigma', type='complex',
                                guilabel='sigma',
                                default=0.0,
                                tip="contuctivity")),
        ('kz', VtableElement('Nz', type='float',
                             guilabel='kz',
                             default=0.0,
                             no_func=True,
                             tip="out-of-plane wave number")),)


def Epsilon_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # - omega^2 * epsilon0 * epsilonr
    fac = -epsilon0 * omega * omega * cnorm
    return SCoeff(exprs, ind_vars, l, g, return_complex=True, scale=fac)


def Sigma_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # v = - 1j * self.omega * v
    fac = - 1j * omega * cnorm
    return SCoeff(exprs, ind_vars, l, g, return_complex=True, scale=fac)


def Mu_Coeff(exprs, ind_vars, l, g, omega,  cnorm):
    # v = mu * v
    fac = mu0/cnorm
    return SCoeff(exprs, ind_vars, l, g, return_complex=True, scale=fac)

def domain_constraints():
   return [EM2D_Vac]

class EM2D_Vac(EM2D_Domain, EM2D_Domain_helper):
    vt = Vtable(data)
    #nlterms = ['epsilonr']

    def get_possible_child(self):
        from .em2d_pml import EM2D_LinearPML
        return [EM2D_LinearPML]

    def has_bf_contribution(self, kfes):
        if kfes == 0:
            return True
        elif kfes == 1:
            return True
        else:
            return False

    def has_mixed_contribution(self):
        return True

    def get_mixedbf_loc(self):
        '''
        r, c, and flag1, flag2 of MixedBilinearForm
           flag1 : take transpose
           flag2 : take conj
        '''
        return [(0, 1, 1, 1), (1, 0, 1, 1)]

    def get_coeffs(self):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s, kz = self.vt.make_value_or_expression(self)

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns
        coeff1 = Epsilon_Coeff([e], ind_vars, l, g, omega, cnorm)
        coeff2 = Mu_Coeff([m], ind_vars, l, g, omega, cnorm)
        coeff3 = Sigma_Coeff([s], ind_vars, l, g, omega, cnorm)

        #dprint1("epsr, mur, sigma " + str(coeff1) + " " + str(coeff2) + " " + str(coeff3))
        return coeff1, coeff2, coeff3, kz

    def get_coeffs_2(self):
        # e, m, s
        coeff1, coeff2, coeff3, kz = self.get_coeffs()

        if self.has_pml():
            coeff2 = self.make_PML_coeff(coeff2)

            '''
            tmp = ComplexMatrixInv(coeff2)
            mu11 = ComplexMatrixSlice(tmp, [0,1], [0,1])
            mu11 = ComplexMatrixAdj(mu11)
            mu22 = ComplexMatrixSlice(tmp, [2], [2])                        
            k2_over_mu11 =   ComplexMatrixProduct(mu11, kz*kz)
            ik_over_mu11 =   ComplexMatrixProduct(mu11, 1j*kz)
            '''
            tmp = coeff2.inv()
            mu11 = tmp[[0, 1], [0, 1]]
            mu11 = mu11.adj()
            mu22 = tmp[2, 2]
            k2_over_mu11 = mu11*(kz*kz)
            ik_over_mu11 = mu11*(1j*kz)
            mu = [mu11, mu22, k2_over_mu11, ik_over_mu11]

            '''
            coeff4 = ComplexSum(coeff1, coeff3)
            '''
            coeff4 = coeff1 + coeff3
            coeff4 = self.make_PML_coeff(coeff4)

            '''
            eps11 = ComplexMatrixSlice(coeff4, [0,1], [0,1])
            eps21 = ComplexMatrixSlice(coeff4, [0,1], [2])
            eps12 = ComplexMatrixSlice(coeff4, [2], [0,1])
            eps22 = ComplexMatrixSlice(coeff4, [2], [2])
            '''
            eps11 = coeff4[[0, 1], [0, 1]]
            eps21 = coeff4[[0, 1], 2]
            eps12 = coeff4[2, [0, 1]]
            eps22 = coeff4[2, 2]

            eps = [eps11, eps12, eps21, eps22]

        else:
            '''
            coeff2 = ComplexPow(coeff2, -1)
            '''
            coeff2 = coeff2**(-1)
            one_over_mu = coeff2
            '''
            k2_over_mu =   ComplexProduct(coeff2, kz*kz)
            ik_over_mu =   ComplexProduct(coeff2, 1j*kz)
            '''
            k2_over_mu = coeff2*(kz*kz)
            ik_over_mu = coeff2*(1j*kz)

            mu = [one_over_mu, one_over_mu, k2_over_mu, ik_over_mu]

            '''
            coeff4 = ComplexSum(coeff1, coeff3)
            '''
            coeff4 = coeff1 + coeff3

            eps = [coeff4, None, None, coeff4]

        return eps, mu, kz

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        #neg_w2eps, one_over_mu, k2_over_mu, ik_over_mu, neg_iwsigma, kz= self.get_coeffs_2(real)

        eps, mu, kz = self.get_coeffs_2()

        self.set_integrator_realimag_mode(real)
        self.call_bf_add_integrator(eps,  mu, kz, engine, a, kfes)

        '''
        if kfes == 0: ## ND element (Epoloidal)
            if real:       
                dprint1("Add ND contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add ND contribution(imag)" + str(self._sel_index))
                
            self.add_integrator(engine, 'mur', mu[1],
                                a.AddDomainIntegrator,
                                mfem.CurlCurlIntegrator)
            self.add_integrator(engine, 'epsilon_sigma', eps[0],
                                a.AddDomainIntegrator,
                                mfem.VectorFEMassIntegrator)
            

            if kz != 0:
                self.add_integrator(engine, 'mur', mu[2],
                                    a.AddDomainIntegrator,
                                    mfem.VectorFEMassIntegrator)
                
        elif kfes == 1: ## ND element (Epoloidal)
            if real:
                dprint1("Add H1 contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add H1 contribution(imag)" + str(self._sel_index))

            self.add_integrator(engine, 'mur_2', mu[0],
                                a.AddDomainIntegrator,
                                mfem.DiffusionIntegrator)
            self.add_integrator(engine, 'epsilonr_sigma', esp[3]
                                a.AddDomainIntegrator,
                                mfem.MassIntegrator)
        else:
            pass
        '''

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real=True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        eps, mu, kz = self.get_coeffs_2()

        self.set_integrator_realimag_mode(real)
        self.call_mix_add_integrator(eps, mu, engine, mbf, r, c, is_trans)

        '''       
        neg_w2eps, one_over_mu, k2_over_mu, ik_over_mu, neg_iwsigma, kz = self.get_coeffs_2(real)
        
        self.set_integrator_realimag_mode(real)
        
        if r == 1 and c == 0:
            # (-a u_vec, div v_scalar)           
            itg =  mfem.MixedVectorWeakDivergenceIntegrator
        else:
            itg =  mfem.MixedVectorGradientIntegrator

        self.add_integrator(engine, 'mur', ik_over_mu,
                                mbf.AddDomainIntegrator, itg)
        '''

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant
        e, m, s, kz = self.vt.make_value_or_expression(self)

        if len(self._sel_index) == 0:
            return

        self.do_add_scalar_expr(v, suffix, ind_vars,
                                'sepsilonr', e, add_diag=3)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'smur', m, add_diag=3)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'ssigma', s, add_diag=3)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')

        add_constant(v, 'kz', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

